#include <string.h>
#include <assert.h>

#include "palm-extras.h"
#include "palmrcp.h"

typedef union _ctl_data {
    struct {
      UInt16 field_id;
      MemHandle field_mh;
    } string;
    struct {
      UInt16 checkbox_id;
    } boolean;
    struct {
      UInt16 trigger_id, list_id;
      char **items;
      int n_items, n_alloc;
    } choice;
} ctl_data;

typedef struct _cfg_data {
  char *title;
  config_item *cfg;     /* array, allocated from midend_get_config. */
  int which;

  FormType *frmP;       /* may change! */
  ctl_data *ctls;       /* array, allocated by us. */
  int next_id, next_top, winheight, winwidth, need_gsi;
  int max_labelw_long, max_labelw_short;
  int col2x_long, col2x_short;
  int multiline, fldlines, pagelines;
} cfg_data;

#define DBOX_BORDER     4       /* from sides and bottom */
#define DBOX_TITLE      13      /* from top */
#define DBOX_DEPTH      12      /* depth of each line */
#define DBOX_GAP        1       /* gap between one row and the next */
#define DBOX_BUTTONH    12      /* height of button */
#define DBOX_BUTTONW    36      /* (desired) width of button */

#define DBOX_LABELH     11
#define DBOX_STRINGH    11
#define DBOX_BOOLEANH   12
#define DBOX_CHOICESH   12

#define CUSTID_START            10100   /* lots more than palmrcp.h */
#define CUSTID_FORM             (CUSTID_START + 0)
#define CUSTID_OK               (CUSTID_START + 1)
#define CUSTID_CANCEL           (CUSTID_START + 2)

#define CUSTID_NEXT             (CUSTID_START + 10)

void *FrmEasyPtr(FormType *frmP, UInt16 id)
{
  UInt16 index = FrmGetObjectIndex(frmP, id);
  assert(index != frmInvalidObjectId);
  return FrmGetObjectPtr(frmP, index);
}

/* --- event handler for multiline string field and scroll bar --- */

/* Code shamelessly adapted from Memopad source code in the SDK. */

static void ConfigUpdateScrollBar(FieldPtr fld, ScrollBarPtr bar) CUSTOM_SECT;
static void ConfigUpdateScrollBar(FieldPtr fld, ScrollBarPtr bar)
{
  UInt16 sclpos, txtheight, fldheight;
  Int16 max;
  
  FldGetScrollValues(fld, &sclpos, &txtheight, &fldheight);
  if (fldheight <= 1) return;
  if (txtheight > fldheight) {
    max = (txtheight - fldheight) + FldGetNumberOfBlankLines(fld);
  }
  else if (sclpos)
    max = sclpos;
  else
    max = 0;
  SclSetScrollBar(bar, sclpos, 0, max, fldheight - 1);
}

static Boolean ConfigHandleEvent(EventPtr event) CUSTOM_SECT;
static Boolean ConfigHandleEvent(EventPtr event)
{
  Boolean handled = false;
  Int16 numlines;
  FieldPtr fld;
  ScrollBarPtr bar;

  switch (event->eType) {
  case fldChangedEvent:
    fld = event->data.fldChanged.pField;
    /* We can also get fldChangedEvent for a single line field
     * without a scroll bar, ignore this */
    if (FldGetVisibleLines(fld) > 1) {
      bar = FrmEasyPtr(FrmGetActiveForm(), frmMlStringScrollBar);
      ConfigUpdateScrollBar(fld, bar);
    }
    break;

  case sclRepeatEvent:
    fld = FrmEasyPtr(FrmGetActiveForm(), frmMlStringField);
    bar = event->data.sclRepeat.pScrollBar;
    numlines = event->data.sclRepeat.newValue - event->data.sclRepeat.value;
    if (numlines < 0)
      FldScrollField(fld, -numlines, winUp);
    else if (numlines > 0)
      FldScrollField(fld, numlines, winDown);
    ConfigUpdateScrollBar(fld, bar);
    break;

  default:
  }
  return handled;
}

/* --- adding controls --- */

#define FIELD_MINSIZE 150

static int AddField(cfg_data *cfg, config_item *cfgi, ctl_data *ctli) CUSTOM_SECT;
static int AddField(cfg_data *cfg, config_item *cfgi, ctl_data *ctli)
{
  MemHandle htmp;
  MemPtr data;
  FieldType *field;
  ScrollBarType *bar = (ScrollBarType *) 0;
  UInt16 fldx, fldwidth;
  /* leave room for at least) twice the existing string length */
  UInt16 fsize = max(strlen(cfgi->sval) * 2, FIELD_MINSIZE);

  debug(("adding new text field ID %d", cfg->next_id));
  ctli->string.field_id = cfg->next_id++;
  if (cfg->multiline) {
    fldx = DBOX_BORDER;
    fldwidth = cfg->winwidth - 2 * DBOX_BORDER - 7;
    cfg->fldlines = FldCalcFieldHeight(cfgi->sval, fldwidth);
    cfg->pagelines = (cfg->fldlines < 10 ? cfg->fldlines : 10);
  } else {
    fldx = cfg->col2x_long;
    fldwidth = cfg->winwidth - cfg->col2x_long - DBOX_BORDER;
    cfg->fldlines = cfg->pagelines = 1;
  }

  if (cfg->multiline)
  {
    RectangleType bounds;

    /* adjust multiline text field */
    field = (FieldType *) FrmEasyPtr(cfg->frmP, frmMlStringField);
    FldSetMaxChars(field, fsize);
    bounds.topLeft.x = fldx;
    bounds.topLeft.y = cfg->next_top;
    bounds.extent.x = fldwidth;
    bounds.extent.y = cfg->pagelines * DBOX_STRINGH;
    FrmSetObjectBounds(cfg->frmP, 
                       FrmGetObjectIndex(cfg->frmP, frmMlStringField),
                       &bounds);
    /* adjust corresponding scroll bar */
    bar = (ScrollBarType *) FrmEasyPtr(cfg->frmP, frmMlStringScrollBar);
    bounds.topLeft.x = fldx + fldwidth;
    bounds.extent.x = 7;
    FrmSetObjectBounds(cfg->frmP,
                       FrmGetObjectIndex(cfg->frmP, frmMlStringScrollBar),
                       &bounds);
  }
  else {
    field =
      FldNewField((void**)&cfg->frmP, ctli->string.field_id,
                fldx, cfg->next_top,
                fldwidth, cfg->pagelines * DBOX_STRINGH,
                stdFont,
                fsize,  /* max chars... */
                true,   /* editable */
                grayUnderline,
                true,   /* single line */
                false,  /* no dynamic size */
                leftAlign,
                true,   /* use auto-shift */
                false,  /* no scroll bar */
                false   /* not strictly numeric */);
  }
  if (!field) {
    debug(("AddField: unable to add field for %s.", cfgi->name));
    return (-1);
  }

  /* scunge city; allocate our own block, copy string
   * into it, replace existing block (if it exists) so
   * we have control. */
  ctli->string.field_mh = MemHandleNew(fsize);
  if (!ctli->string.field_mh) fatal_malloc(fsize); /* OOM */

  data = MemHandleLock(ctli->string.field_mh);
  StrCopy(data, cfgi->sval);
  MemHandleUnlock(ctli->string.field_mh);

  htmp = FldGetTextHandle(field);
  FldSetTextHandle(field, ctli->string.field_mh);
  if (htmp) MemHandleFree(htmp);

  if (cfg->multiline) {
    /* set scroll bar parameters */
    UInt16 scrlpos, txtheight, fldheight, max;
    FldGetScrollValues(field, &scrlpos, &txtheight, &fldheight);
    if (txtheight > fldheight)
      max = txtheight - fldheight;
    else if (scrlpos)
      max = scrlpos;
    else
      max = 0;
    SclSetScrollBar(bar, scrlpos, 0, max, fldheight - 1);
  }

  return 0;
}

static int AddCheckbox(cfg_data *cfg,
                       config_item *cfgi, ctl_data *ctli) CUSTOM_SECT;
static int AddCheckbox(cfg_data *cfg,
                       config_item *cfgi, ctl_data *ctli)
{
  ControlType *checkbox;

  debug(("adding new checkbox ID %d", cfg->next_id));
  ctli->boolean.checkbox_id = cfg->next_id++;
  checkbox =
    CtlNewControl((void**)&cfg->frmP, ctli->boolean.checkbox_id,
                  checkboxCtl, "",
                  cfg->col2x_long, cfg->next_top,
                  DBOX_BOOLEANH, DBOX_BOOLEANH,
                  boldFont /* ignored */, 0, true);
  if (!checkbox) {
    debug(("AddCheckbox: unable to add checkbox for %s.", cfgi->name));
    return -1;
  }
  CtlSetValue(checkbox, (cfgi->ival != 0) ? 1 : 0);

  return 0;
}

static int AddPopup(cfg_data *cfg,
                    config_item *cfgi, ctl_data *ctli) CUSTOM_SECT;
static int AddPopup(cfg_data *cfg,
                    config_item *cfgi, ctl_data *ctli)
{
  char c, *p, *q, *str;
  ListType *listP;
  ControlType *trigger;

  debug(("adding new popup-trigger ID %d", cfg->next_id));
  ctli->choice.trigger_id = cfg->next_id++;
  trigger =
    CtlNewControl((void**)&cfg->frmP, ctli->choice.trigger_id,
                  popupTriggerCtl, "some text",
                  cfg->col2x_short, cfg->next_top,
                  cfg->winwidth - cfg->col2x_short - DBOX_BORDER, DBOX_CHOICESH,
                  stdFont, 0, true);
  if (!trigger) {
    debug(("AddPopup: unable to add trigger for %s.", cfgi->name));
    return -1;
  }

  debug(("adding new popup-list ID %d", cfg->next_id));
  ctli->choice.list_id = cfg->next_id++;
  if (LstNewList((void**)&cfg->frmP, ctli->choice.list_id,
                cfg->col2x_short, cfg->next_top,
                cfg->winwidth - cfg->col2x_short - DBOX_BORDER, DBOX_CHOICESH,
                stdFont, 3, ctli->choice.trigger_id)) {
    debug(("AddPopup: unable to add list for %s.", cfgi->name));
    return -1;
  }
  listP = FrmEasyPtr(cfg->frmP, ctli->choice.list_id);
  if (!listP) {
    debug(("AddPopup: unable to fetch list ptr (!) for %s.", cfgi->name));
    return -1;
  }

  /* mostly nicked from the windows one. However, here we have to
   * allocate a list-of-char* for the list ctrl, as well as keeping
   * all our strings around until we finish. These are all
   * deallocated at the end of get_config. */
  ctli->choice.n_alloc = 10;
  ctli->choice.items = snewn(ctli->choice.n_alloc, char *);
  ctli->choice.n_items = 0;

  p = cfgi->sval;
  c = *p++;
  while (*p) {
    q = p;
    while (*q && *q != c) q++;
    str = snewn(q-p+1, char);
    strncpy(str, p, q-p);
    str[q-p] = '\0';
    ctli->choice.items[ctli->choice.n_items++] = str;
    if (ctli->choice.n_items >= ctli->choice.n_alloc) {
      ctli->choice.n_alloc *= 2;
      ctli->choice.items = sresize(ctli->choice.items, ctli->choice.n_alloc, char *);
    }
    if (*q) q++;
    p = q;
  }
  LstSetListChoices(listP, ctli->choice.items, ctli->choice.n_items);
  LstSetHeight(listP, ctli->choice.n_items);
  LstSetSelection(listP, cfgi->ival);
  debug(("current list selection %d", LstGetSelection(listP)));

  /* get the ptr again, it might have changed */
  trigger = FrmEasyPtr(cfg->frmP, ctli->choice.trigger_id);
  CtlSetLabel(trigger, ctli->choice.items[cfgi->ival]);
  debug(("current trigger label '%s'", CtlGetLabel(trigger)));

  return 0;
}

/* --- adding labels --- */

#define LABEL_FONT boldFont
#define MAX_LABELW_LONG 110 /* for string, boolean */
#define MAX_LABELW_SHORT 80 /* for choices */

/* scunge. why don't labels have colons themselves? */
static const char *ToLabel(const char *label) CUSTOM_SECT;
static const char *ToLabel(const char *label)
{
  static char l2[50];
  strncpy(l2, label, 48);
  l2[48] = '\0';
  strcat(l2, ":");
  return l2;
}

static int LabelWidth(const char *text) CUSTOM_SECT;
static int LabelWidth(const char *text)
{
  FontID oldfont = FntSetFont(LABEL_FONT);
  int width = FntCharsWidth(text, strlen(text));
  FntSetFont(oldfont);
  debug(("LabelWidth: width of '%s' is %d.", text, width));
  return width;
}

static int AddLabelR(FormType **frmPP, config_item *ctli,
                      cfg_data *cfg) CUSTOM_SECT;
static int AddLabelR(FormType **frmPP, config_item *ctli,
                      cfg_data *cfg)
{
  const char *label = ToLabel(ctli->name);
  int labelw = LabelWidth(label), labell = DBOX_BORDER;
  int top = cfg->next_top;
  int maxlw = (ctli->type == C_CHOICES) ?
    cfg->max_labelw_short : cfg->max_labelw_long;

  debug(("adding label '%s' ID %d", label, cfg->next_id));
  if (ctli->type == C_STRING && cfg->multiline) {
    cfg->next_top += DBOX_LABELH;
    cfg->winheight += DBOX_LABELH;
    if (labelw < maxlw)
      labell += (maxlw - labelw) / 2;
  } else {
    if (labelw > maxlw)
      cfg->next_top += DBOX_LABELH;
    else
      labell += maxlw - labelw;
  }
  if (!FrmNewLabel(frmPP, cfg->next_id++, label, labell, top, LABEL_FONT))
    return -1;
  return 0;
}

static void AddLabelWidth(config_item *ctli, cfg_data *cfg) CUSTOM_SECT;
static void AddLabelWidth(config_item *ctli, cfg_data *cfg)
{
  int maxw = 0, *curr_maxw = NULL, labelw = LabelWidth(ToLabel(ctli->name));

  switch (ctli->type) {
  case C_STRING:
  case C_BOOLEAN:
    maxw = MAX_LABELW_LONG;
    curr_maxw = &cfg->max_labelw_long;
    break;

  case C_CHOICES:
    maxw = MAX_LABELW_SHORT;
    curr_maxw = &cfg->max_labelw_short;
    break;

  default:
    assert("shouldn't get here!");
  }
  if (labelw > maxw)
    cfg->winheight += DBOX_LABELH;
  else if (labelw > *curr_maxw)
    *curr_maxw = labelw;
}

/* --- adding buttons --- */

static int AddButton(FormType **frmPP, int id, const char *text,
                     cfg_data *cfg, int left) CUSTOM_SECT;
static int AddButton(FormType **frmPP, int id, const char *text,
                     cfg_data *cfg, int left)
{
  return (CtlNewControl((void**)frmPP, id, buttonCtl, text,
                     left, cfg->winheight - DBOX_BUTTONH - 5,
                     DBOX_BUTTONW, DBOX_BUTTONH,
                     stdFont, 0, true)) ? 0 : -1;
}

/* --- putting the whole thing together --- */

enum {
  DLG_SIZE,             /* returns no. of controls */
  DLG_CREATE,           /* returns non-zero on error */
  DLG_READ,             /* returns 0 */
  DLG_DESTROY           /* returns 0 */
};

static int enum_config(cfg_data *cfg, int stage) CUSTOM_SECT;
static int enum_config(cfg_data *cfg, int stage)
{
  config_item *ctli;
  int i, j, ret = 0;

  for (ctli = cfg->cfg, i = 0; ctli->type != C_END; ctli++, i++) {
    ctl_data *ctld = &cfg->ctls[i];

    switch (stage) {
    case DLG_SIZE:
      AddLabelWidth(ctli, cfg);
      switch (ctli->type) {
      case C_STRING:
        cfg->winheight += DBOX_STRINGH;
        cfg->need_gsi = 1;
        break;

      case C_BOOLEAN:
        cfg->winheight += DBOX_BOOLEANH;
        break;

      case C_CHOICES:
        cfg->winheight += DBOX_CHOICESH;
        break;
      }
      cfg->winheight += DBOX_GAP;
      ret++;
      break;

    case DLG_CREATE:
      if (AddLabelR(&cfg->frmP, ctli, cfg)) return -1;
      switch (ctli->type) {
      case C_STRING:
        if (AddField(cfg, ctli, ctld)) return -1;
        cfg->next_top += cfg->pagelines * DBOX_STRINGH;
        if (cfg->pagelines > 1)
          cfg->winheight += (cfg->pagelines - 1) * DBOX_STRINGH;
        break;

      case C_BOOLEAN:
        if (AddCheckbox(cfg, ctli, ctld)) return -1;
        cfg->next_top += DBOX_BOOLEANH;
        break;

      case C_CHOICES:
        if (AddPopup(cfg, ctli, ctld)) return -1;
        cfg->next_top += DBOX_CHOICESH;
        break;
      }
      cfg->next_top += DBOX_GAP;
      break;

    case DLG_READ:
      switch (ctli->type) {
      case C_STRING:
        sfree(ctli->sval);
        ctli->sval = dupstr(MemHandleLock(ctld->string.field_mh));
        MemHandleUnlock(ctld->string.field_mh);
        break;

      case C_BOOLEAN:
        {
          ControlType *ctlP = FrmEasyPtr(cfg->frmP,ctld->boolean.checkbox_id);
          ctli->ival = CtlGetValue(ctlP);
        }
        break;

      case C_CHOICES:
        {
          ListType *listP = FrmEasyPtr(cfg->frmP, ctld->choice.list_id);
          ctli->ival = LstGetSelection(listP);
        }
        break;
      }
      break;

    case DLG_DESTROY:
      switch (ctli->type) {
      case C_CHOICES:
        for (j = 0; j < ctld->choice.n_items; j++)
          sfree(ctld->choice.items[j]);
        sfree(ctld->choice.items);
        break;
      }
      break;

    default:
      fatal("enum_config unknown enum %d", stage);
    }
  }
  return ret;
}

/* returns non-zero on error. */
int get_config(frontend *fe, int which)
{
  cfg_data cfg;
  int n_ctls, ret = -1;
  RectangleType drawsize, frmsize;

  /* I think there's a bug in DIA for modal forms until OS6.
   * I get a 'bitmap.c, Line:244, Null width or height' error
   * if I create this where it should be on OS5. */
  /* WinGetDrawWindowBounds(&drawsize);*/
  drawsize.topLeft.x = drawsize.topLeft.y = 0;
  drawsize.extent.x = drawsize.extent.y = 160;

  memset(&cfg, 0, sizeof(cfg_data));
  cfg.which = which;
  if (which >= 0)
    cfg.cfg = midend_get_config(fe_me(fe), which, &(cfg.title));
  else
    cfg.cfg = palm_get_config(fe, which, &(cfg.title));

  if (which == CFG_SEED || which == CFG_DESC)
    cfg.multiline = 1;
  else
    cfg.multiline = 0;

  cfg.winheight = DBOX_TITLE;
  cfg.winwidth = drawsize.extent.x - 4;
  n_ctls = enum_config(&cfg, DLG_SIZE);
  cfg.winheight += (DBOX_BUTTONH + DBOX_BORDER + DBOX_BORDER);

  debug(("get_config: %d cfgs, winheight = %d, %s-line strings",
         n_ctls, cfg.winheight, cfg.multiline ? "multi" : "single"));
  cfg.ctls = snewn(n_ctls, ctl_data);

  /* generate form itself */

  if (cfg.multiline)
  {
    cfg.frmP = FrmInitForm(frmMlString);
    if (cfg.frmP) FrmSetTitle(cfg.frmP, cfg.title);
  }
  else {
    cfg.frmP = FrmNewForm(CUSTID_START, cfg.title,
                    2, drawsize.extent.y-2-cfg.winheight,
                    cfg.winwidth, cfg.winheight,
                    true, CUSTID_CANCEL, 0, editMenu);
  }
  if (!cfg.frmP) {
    FrmAlert(formFailed);
    return (-1);
  }
  /* disabled for same reason. */
  /* SetupDIA(cfg.frmP);*/

  cfg.next_id = CUSTID_NEXT;
  cfg.col2x_long  = DBOX_BORDER + cfg.max_labelw_long  + DBOX_BORDER;
  cfg.col2x_short = DBOX_BORDER + cfg.max_labelw_short + DBOX_BORDER;
  cfg.next_top = DBOX_TITLE;
  if (dia_features) {
    /* Calling FrmDoDialog below will force the DIA to open. We open it
     * here already to get things like selection list height computed
     * correctly */
    PINSetInputAreaState(pinInputAreaOpen);
  }
  if (enum_config(&cfg, DLG_CREATE)) goto dialog_bad;

  /* add OK/Cancel buttons, and the GSI if we want it. */
  if (AddButton(&cfg.frmP, CUSTID_OK,     "OK", &cfg, 5) ||
      AddButton(&cfg.frmP, CUSTID_CANCEL, "Cancel", &cfg,
                5 + DBOX_BUTTONW + 6))
    goto dialog_bad;

  if (cfg.need_gsi && !FrmNewGsi(&cfg.frmP, cfg.winwidth - 17, cfg.winheight - 13))
    goto dialog_bad;

  /* Adjust form boundaries if we have multiple lines */
  FixupWinGetBounds(FrmGetWindowHandle(cfg.frmP), &frmsize);
  if (frmsize.extent.y != cfg.winheight) {
    UInt16 offsety;
    offsety = cfg.winheight - frmsize.extent.y;
    frmsize.topLeft.y -= offsety;
    frmsize.extent.y += offsety;
    WinSetBounds(FrmGetWindowHandle(cfg.frmP), &frmsize);
  }

  /* Check form does not go off screen. Since it is modal, max height is 156 */
  assert(frmsize.extent.y <= 156);

  if (nav_features == 1) {
    /* Set navigation state to Interaction Mode & App Focus Mode */
    FrmSetNavState(cfg.frmP, 0);
  }

  /* display (and process form) */
  FrmSetEventHandler(cfg.frmP, ConfigHandleEvent);
#ifdef HAVE_HANDERA_SDK
  if (handera && cfg.multiline) VgaFormModify(cfg.frmP, vgaFormModify160To240);
#endif

dialog:
  if (FrmDoDialog(cfg.frmP) == CUSTID_OK) {
    char *err;

    assert(!enum_config(&cfg, DLG_READ));
    if (which >= 0)
      err = midend_set_config(fe_me(fe), which, cfg.cfg);
    else
      err = palm_set_config(fe, which, cfg.cfg);
    if (err) {
      FrmCustomAlert(validationError, err, NULL, NULL);
      goto dialog;
    } else {
      ret = 0;
    }
  } else {
    if (which < 0) palm_free_config(fe, which, cfg.cfg);
  }

cleanup:
  FrmDeleteForm(cfg.frmP);
  assert(!enum_config(&cfg, DLG_DESTROY));
  free_cfg(cfg.cfg);
  sfree(cfg.ctls);
  sfree(cfg.title);
  if (dia_features) {
    /* Undo the effect of forcing the DIA open above */
    PINSetInputAreaState(pinInputAreaUser);
  }
  return ret;

dialog_bad:
  FrmAlert(formFailed);
  goto cleanup;
}

