/* This was lovingly ripped off from Jim Rees' 'httpd' software,
 * available from http://www.citi.umich.edu/u/rees/pilot/#httpd
 *
 * The license is 'postcardware': from his web site:
 *
 * "All of the software on this page is postcardware. It is free in both senses of
 * the word, but if you use it and like it I would appreciate getting a postcard
 * from you. My address is:
 *
 * Jim Rees
 * Center for Information Technology Integration
 * 535 W William
 * Ann Arbor, Michigan 48103 USA
 *
 * If you really like my software and are feeling extraordinarily generous, you
 * might consider buying me something from my Amazon wish list 
 * [http://www.amazon.com/gp/registry/registry.html/?id=32RL340DLOI2Q&items-per-page=50&sort=priority&filter=all&reveal=all]."
 *
 * I mainly made it use va_args correctly and added '%n' support.
 */

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "puzzles.h"

#define	SPC	01
#define	STP	02

#define	SHORT	0
#define	REGULAR	1
#define	LONG	2
#define	INT	0
#define	FLOAT	1

#define iop_eof	(-1)

typedef struct _iop {
    const char *p;
    int r;
} iop_s;

#define iop_getc(f) (--(f)->r < 0 ? iop_eof : *(f)->p++)
#define iop_ungetc(c, f) (++(f)->r, *--(f)->p)

//#define isdigit(c)	(c >= '0' && c <= '9')

static char _sctab[256] = {
    0,0,0,0,0,0,0,0,
    0,SPC,SPC,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
    SPC,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
};

static int vsscanf(const char *s, const char *fmt, va_list args) PUZZ_SECT;
static int _innum(int hadptr, va_list *args, int type, int len,
                  int size, iop_s *iop, int *eofptr) PUZZ_SECT;
static int _instr(char *ptr, int type, int len, iop_s *iop, int *eofptr) PUZZ_SECT;
static char *_getccl(const char *s);

#ifdef TEST
main()
{
    char s1[8], s2[8];
    int i, j, nmatched;
    long l;

    sprintf(s1, "");
    sprintf(s2, "");
    i = j = 0;
    l = 0;

    printf("s1=0x%x, s2=0x%x, &i=0x%x, &j=0x%x, &l=0x%x\n",
           s1, s2, &i, &j, &l);

    nmatched = sscanf("foo barx 1 2 a", "%s %[^x]x %d %ld %x", s1, s2, &i, &l, &j);
    printf("nmatched = %d\n'%s' '%s' '%d' '%ld' '%x'\n", nmatched, s1, s2, i, l, j);
    exit(0);
}
#endif

int sscanf(const char *s, const char *fmt, ...)
{
    va_list args;
    int l;

    debug(("sscanf, scanning '%s' for '%s", s, fmt));

    va_start(args, fmt);
    l = vsscanf(s, fmt, args);
    va_end(args);
    debug(("sscanf scanned '%s' for '%s' and found %d matches",
           s, fmt, l));
    return l;
}

int vsscanf(const char *s, const char *fmt, va_list args)
{
    register int ch;
    int nmatch = 0, len, ch1;
    int fileended = 0, size, hadptr, suyb;
    const char *sloop = s;
    iop_s iop;

    iop.p = s;
    for (iop.r = 0; *sloop++; iop.r++)
	;

    while (1) {
	switch (ch = *fmt++) {
	case '\0': 
	    return (nmatch);
	case '%': 
	    if ((ch = *fmt++) == '%')
		goto def;
            if (ch == 'n') {
                int *cptr = va_arg(args, int *);
                *cptr = (int)(iop.p - s);
                break;
            }
            hadptr = 0;
	    if (ch != '*')
                hadptr = 1;
            else
		ch = *fmt++;
	    len = 0;
	    size = REGULAR;
	    while (isdigit(ch)) {
		len = len*10 + ch - '0';
		ch = *fmt++;
	    }
	    if (len == 0)
		len = 30000;
	    if (ch == 'l') {
		size = LONG;
		ch = *fmt++;
	    } else if (ch == 'h') {
		size = SHORT;
		ch = *fmt++;
	    } else if (ch == '[')
		fmt = _getccl(fmt);
	    if (ch >= 'A' && ch <= 'Z') {
		ch |= 0x20;
		size = LONG;
	    }
	    if (ch == '\0')
		return(-1);
            if (fileended)
		return(nmatch ? nmatch : -1);
	    if (_innum(hadptr, &args, ch, len, size, &iop, &fileended) && hadptr)
		nmatch++;
	    break;

	case ' ':
	case '\n':
	case '\t': 
	    while ((ch1 = iop_getc(&iop)) == ' ' || ch1 == '\t' || ch1 == '\n')
		;
	    if (ch1 != iop_eof)
		suyb = iop_ungetc(ch1, &iop);
	    break;

	default: 
	def:
	    ch1 = iop_getc(&iop);
	    if (ch1 != ch) {
		if (ch1 == iop_eof)
		    return -1;
		suyb = iop_ungetc(ch1, &iop);
		return(nmatch);
	    }
	}
    }
}

static int
_innum(int hadptr, va_list *args, int type, int len,
       int size, iop_s *iop, int *eofptr)
{
  register char *np;
  char numbuf[64];
  register int c, base;
  int expseen = 0, scale = INT, negflg = 0, c1, ndigit = 0;
  long lcval = 0;

  if (type == 'c' || type == 's' || type == '[') {
    char *cptr = hadptr ? va_arg(*args, char *) : NULL;
    return(_instr(cptr, type, len, iop, eofptr));
  }
#ifdef PRINT_FLOATS
  if (type == 'e' || type == 'f')
    scale = FLOAT;
#endif
  if (type == 'o')
    base = 8;
  else if (type == 'x')
    base = 16;
  else
    base = 10;
  np = numbuf;
  while ((c = iop_getc(iop)) == ' ' || c == '\t' || c == '\n')
    ;
  if (c == '-') {
    negflg++;
    *np++ = c;
    c = iop_getc(iop);
    len--;
  } else if (c == '+') {
    len--;
    c = iop_getc(iop);
  }
  for ( ; --len >= 0; *np++ = c, c = iop_getc(iop)) {
    if (isdigit(c) || (base == 16 && (('a' <= c && c <= 'f') || ('A' <= c && c <= 'F')))) {
      ndigit++;
      if (base==8)
        lcval <<=3;
      else if (base==10)
        lcval = ((lcval<<2) + lcval)<<1;
      else
        lcval <<= 4;
      c1 = c;
      if (isdigit(c))
        c -= '0';
      else if ('a' <= c && c <= 'f')
        c -= 'a' - 10;
      else
        c -= 'A' - 10;
      lcval += c;
      c = c1;
      continue;
    } else if (c == '.') {
      if (base != 10 || scale == INT)
        break;
      ndigit++;
      continue;
    } else if ((c == 'e' || c == 'E') && !expseen) {
      if (base != 10 || scale == INT || ndigit == 0)
        break;
      expseen++;
      *np++ = c;
      c = iop_getc(iop);
      if (c != '+' && c != '-' && ('0' > c || c > '9'))
        break;
    } else
      break;
  }
  if (negflg)
    lcval = -lcval;
  if (c != iop_eof) {
    int suyb = iop_ungetc(c, iop);
    suyb = suyb; /* no, really, I mean it. */
    *eofptr = 0;
  } else
    *eofptr = 1;
  if (!hadptr || np == numbuf || (negflg && np == numbuf+1))
    return 0;
  *np++ = 0;
  switch ((scale << 4) | size) {

#ifdef PRINT_FLOATS
  case (FLOAT<<4) | SHORT:
  case (FLOAT<<4) | REGULAR:
    {
      float *fptr = va_arg(*args, float *);
      *fptr = atof(numbuf);
    }
    break;

  case (FLOAT<<4) | LONG:
    {
      double *dptr = va_arg(*args, double *);
      *dptr = atof(numbuf);
    }
    break;
#endif

  case (INT<<4) | SHORT:
    {
      short *sptr = va_arg(*args, short *);
      *sptr = (short)lcval;
    }
    break;

  case (INT<<4) | REGULAR:
    {
      int *iptr = va_arg(*args, int *);
      *iptr = (int)lcval;
    }
    break;

  case (INT<<4) | LONG:
    {
      long *lptr = va_arg(*args, long *);
      *lptr = lcval;
    }
    break;
  }
  return 1;
}

static int
_instr(char *ptr, int type, int len, iop_s *iop, int *eofptr)
{
    int ch, ignstp;
    char *optr = ptr;

    *eofptr = 0;
    if (type == 'c' && len == 30000)
	len = 1;
    ignstp = (type == 's') ? SPC : 0;
    while ((ch = iop_getc(iop)) != iop_eof && (_sctab[ch] & ignstp))
	;
    if (type == 'c')
	ignstp = 0;
    else if (type == '[')
	ignstp = STP;
    else
	ignstp = SPC;
    while (ch != iop_eof && !(_sctab[ch] & ignstp)) {
	if (ptr)
	    *ptr++ = ch;
	if (--len <= 0)
	    break;
	ch = iop_getc(iop);
    }
    if (ch != iop_eof) {
	if (len > 0) {
	    int suyb = iop_ungetc(ch, iop);
            suyb = suyb;
        }
	*eofptr = 0;
    } else
	*eofptr = 1;
    if (ptr && ptr != optr) {
	if (type != 'c')
	    *ptr++ = '\0';
	return 1;
    }
    return 0;
}

static char *
_getccl(const char *s)
{
    int c, t;

    t = 0;
    if (*s == '^') {
	t++;
	s++;
    }
    for (c = 0; c < (sizeof _sctab / sizeof _sctab[0]); c++)
	if (t)
	    _sctab[c] &= ~STP;
	else
	    _sctab[c] |= STP;
    if ((c = *s) == ']' || c == '-') { /* first char is special */
	if (t)
	    _sctab[c] |= STP;
	else
	    _sctab[c] &= ~STP;
	s++;
    }
    while ((c = *s++) != ']') {
	if (!c)
	    return((char *)--s);
	else if (c == '-' && *s != ']' && s[-2] < *s) {
	    for (c = s[-2] + 1; c < *s; c++)
		if (t)
		    _sctab[c] |= STP;
		else
		    _sctab[c] &= ~STP;
	} else if (t)
	    _sctab[c] |= STP;
	else
	    _sctab[c] &= ~STP;
    }
    return((char *)s);
}
