#! /usr/bin/env python

import os
import sys
import re
import time

def die(s):
  print >>sys.stderr, s
  sys.exit(1)

def mksect(fstem):
  return "g_%s" % (fstem[0:6])

targets = (
    ('midend_.*',       'midend'),
    ('.*dsf.*',         'puzz'),
    ('.*_timer',        'puzz'),
    ('.*blitter.*',     'draw'),
    ('draw_rect_outline','puzz'),
    ('draw.*',          'draw'),
    ('.*_draw',         'draw'),
    ('get_random_seed', 'puzz'),
    ('frontend_.*',     'puzz'),
    ('.*clip',          'draw'),
    ('status_bar',      'draw'),
    ('print_.*',        'draw'),
    ('midend.*timer',   'puzz'),
    ('domino.*',        'puzz'),
    ('free_cfg',        'puzz'),
    ('obfuscate_bitmap','puzz'),
    ('bin2hex',         'puzz'),
    ('hex2bin',         'puzz'),
    ('shuffle',         'puzz'),
    ('move_cursor',     'puzz'),
    ('c2pos',           'puzz'),
    ('c2diff',          'puzz'),
    ('pos2c',           'puzz'),
    ('random_.*',       'puzz'),
    ('SHA_.*',          'puzz'),
    ('.*_combi',        'puzz'),
    ('.*mkhighlight.*', 'puzz'),
    ('text_fallback',   'draw'),
)

sects = {}

def whichsect(fnname):
  for (patt, sect) in targets:
    if re.match('^'+patt+'$', fnname) is not None:
      return sect
  return None

def fixup(mo):
  (decl, fnname) = mo.group(1,2)

  sect = whichsect(fnname)

  sects[fnname] = sect
  if sect == 'puzz':
    decorator = ' PUZZ_SECT'
  elif sect == 'draw':
    decorator = ' DRAW_SECT'
  elif sect == 'custom':
    decorator = ' CUSTOM_SECT'
  elif sect is not None:
    decorator = " GEN_SECT(%s)" % (sect)
  else:
    decorator = ''
    sects[fnname] = '(no section)'

  # decorator = "__attribute__ ((section (\"g_%s\")));" % (sect)

  return "%s%s;" % (decl, decorator)

if __name__ == '__main__':
  lines = sys.stdin.read() # whole file at once.
  sys.stdin.close()

  lines = re.sub(r'(?m)^(\w+[^({=]+?([\w_]+)\([^{]+?\));', fixup, lines)

  sys.stdout.write(lines)
  print >>sys.stdout, "\n"
  print >>sys.stdout, "/***********************************************";
  print >>sys.stdout, " * PalmOS section information added by headersect.py";
  print >>sys.stdout, " * " + time.asctime(time.gmtime(time.time()))
  print >>sys.stdout, " *"
  for fnname in sorted(sects.keys()):
    print >>sys.stdout, " *   %s\t--> %s" % (fnname, sects[fnname])
  print >>sys.stdout, " ***********************************************/";
  print >>sys.stdout, "\n"
  sys.exit(0)

