#ifndef PALM_EXTRAS_H
#define PALM_EXTRAS_H

#include <stdarg.h>
#include "puzzles.h"

#ifdef HAVE_HANDERA_SDK
#include "Vga.h"
#endif

#define PRESET_DEFAULT  (-1)
#define PRESET_CUSTOM   (-2)

typedef struct pergame_data {
  midend *me;
} pergame_data;

struct frontend {
  int currgame;         /* index of game we're playing now. */
  pergame_data *pergame;
  int x, y;             /* game size from midend, in real pixels. */
  UInt32 wscr, hscr;    /* screen size, in OS units (i.e. usually both 160) */

#define FF_UPDATEBMP    1
#define FF_UPDATEST     2
#define FF_TIMERON      4
#define FF_NEWNOCANCEL  8
#define FF_WANTSTORE    16
#define FF_GAMELOCKED   32
#define FF_DOWNPENDING  64
#define FF_ISRIGHT      128
#define FF_CLEANWAIT    256
#define FF_NONEWGAME    512
  unsigned long flags;
  int x1up, y1up;       /* required update region... */
  int x2up, y2up;       /* ...from draw_update. */

  int ncolours, depth;  /* from midend */
  ColorTableType *ct;   /* malloced */
  RGBColorType *colours;/* ptr into ct */
  IndexedColorType *idxs; /* malloced */
  int gcol, tcol;       /* curr. graphics and text colours */

  blitter *bl;
  int xoff, yoff;       /* main form bmp display offset */
  float sc;             /* scaling factor from midend --> bitmap */

  WinHandle old_draw;
  UInt32 tt;            /* time value stored for timer */

  char *statusmsg;
  int statush;          /* height of status bar in pixels. */

  UInt32 down_tt;       /* stored at button push, alongside */
  UInt16 down_x, down_y;/* DOWNPENDING and ISRIGHT. */
};

extern frontend fe_global;

/* --- from palm.c --- */

void FixupWinGetBounds(WinHandle winH, RectangleType *rP) DRAW_SECT;

extern int dia_features, nav_features;
extern int clie, handera;

void SetupDIA(FormPtr frmP) PUZZ_SECT;

midend *fe_me(frontend *fe);
struct pergame_data *fe_game(frontend *fe);

#define CFG_PALM_RCLICK (-1)

config_item *palm_get_config(frontend *fe, int which, char **wintitle) PUZZ_SECT;
char *palm_set_config(frontend *fe, int which, config_item *cfg) PUZZ_SECT;
void palm_free_config(frontend *fe, int which, config_item *cfg) PUZZ_SECT;

int app_change_game(frontend *fe, int gameno) PUZZ_SECT;

/* for circle.c */
void palm_draw_rect(void *handle, int x, int y, int w, int h, int colour) DRAW_SECT;

/* --- from trifle.c --- */

void WinFillTriangle(Coord x0, Coord y0,
                     Coord x1, Coord y1,
                     Coord x2, Coord y2) DRAW_SECT;

void WinFillPolygon(int *points, int npoints, float xsc, float ysc) DRAW_SECT;

/* --- from palm-custom.c --- */

void *FrmEasyPtr(FormType *frmP, UInt16 id) CUSTOM_SECT;
int get_config(frontend *fe, int which) CUSTOM_SECT;

/* --- from pdouble.c --- */

void printDouble(char *s, double x, int prec) DRAW_SECT;

/* --- from vsprintf-palm.c --- */

int snprintf(char *buf, size_t len, const char *fmt, ...);
int vsnprintf(char *buf, size_t len, const char *fmt, va_list args);

/* --- from circle.c --- */

void palm_draw_circle(void *handle, int cx, int cy, int radius,
                      int fillcolour, int outlinecolour) DRAW_SECT;

/* --- from saves.c --- */

struct _options {
  int rclick_millis, rclick_offset;
};

extern struct _options *options;

char *CheckOptions(struct _options *opt) GEN_SECT(saves);

int ReadPrefs(void) GEN_SECT(saves);
void WritePrefs(void) GEN_SECT(saves);

char *midend_im_deserialise(midend *me, unsigned char *ptr, size_t len) PUZZ_SECT;
void midend_im_serialise(midend *me, unsigned char **ptr_r, size_t *len_r) PUZZ_SECT;

/* --- from selectgame.c --- */

Boolean ListHandleEvent(EventPtr event) CUSTOM_SECT;

/* --- from fiveway.c --- */

typedef enum { Up = 0, Down, Left, Right, Center } Direction;

Boolean fiveway_key(EventPtr eventP, Direction *dir) CUSTOM_SECT;

#endif
