/*
 * Note to self: various palm compiler bugs I've run into.
 *
 * http://sourceforge.net/tracker/index.php?func=detail&aid=221845&group_id=4429&atid=104429
 *   [ const object + no optimisation --> crash ]
 *   explains why we have to tread carefully around all of the presets struct lists.
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <PalmOSGlue.h>

#include "palmrcp.h"
#include "palmrcp-magic.h"
#include "puzzles.h"
#include "palm-extras.h"
/* #include "palmOneNavigator.h" */

#ifdef HAVE_HANDERA_SDK
#include "Vga.h"
#endif

#ifdef HAVE_CLIE_SDK
#include "SonyCLIE.h"
UInt16 hrRefNum;
#endif

/* OS capabilities of the system we're running on. */
int hd_features, dia_features, nav_features, rgb_colour;
int handera, clie;
double scr_density; /* Multiplication factor from 160x160. */
UInt16 tps;         /* ticks-per-second */
UInt32 romv, kdensity;

struct blitter {
  BitmapType *bmp_off, *bmp_half;
  WinHandle win_off;
  int w, h; /* size (unscaled) */
  int x, y; /* coords saved from (unscaled) */
};

midend *fe_me(frontend *fe) {
  return fe->pergame[fe->currgame].me;
}

struct pergame_data *fe_game(frontend *fe) {
  return &(fe->pergame[fe->currgame]);
}

/* converts midend coords to bitmap pixels... */
#define SCALE(x) ((x)*(fe->sc))
/* ...and the other way. */
#define RSCALE(x) ((x)/(fe->sc))

frontend fe_global;

typedef struct fontinfo {
  FontID fid;
  MemHandle fonth; /* only for our own. */
  FontType *fontp;  /* ditto. */
  int ysize;
} fontinfo;

#define NFONTS_SYS  2
#define NFONTS_LORES 4 /* MUST start at 128! */
#define NFONTS_HIRES 4 /* MUST follow LORES. */
#define NFONTS (NFONTS_SYS + NFONTS_LORES + NFONTS_HIRES)
static fontinfo fontinfos[NFONTS];

/* --- system fixup fns --- */

/* scunge scunge scunge to work around the weirdy
 * qsort in palmos. */
static int (*qptr)(const void *, const void *);

static Int16 pcomp(void *p1, void *p2, Int32 ctx) PUZZ_SECT;
static Int16 pcomp(void *p1, void *p2, Int32 ctx)
{
  return qptr(p1, p2);
}

void qsort(void *base, UInt16 n, Int16 sz,
           int (*compare)(const void *, const void *))
{
  qptr = compare;
  SysQSort(base, n, sz, pcomp, 0);
}

double fabs(double x)
{
  return (x < 0.0) ? (-x) : (x);
}

double atof(const char *s)
{
  FlpCompDouble fd;
  FlpBufferAToF(&(fd.fd), s);
  return fd.d;
}

/*int sscanf(const char *buf, const char *arg, ...)
{
  return EOF;
}*/

char *getenv(const char *env)
{
  return NULL;
}

void FixupWinGetBounds(WinHandle winH, RectangleType *rP)
{
  /* fixup for pre-OS4 machines. */
  WinHandle oldH = WinSetDrawWindow(winH);
  WinGetDrawWindowBounds(rP);
  WinSetDrawWindow(oldH);
}

void palm__assert(int cond, const char *file, int line, const char *cond_s)
{
  if (!cond) ErrDisplayFileLineMsg(file, (UInt16)line, cond_s);
}

/* --- system-specific fns --- */

void fatal(char *fmt, ...)
{
  char buf[2048];
  va_list ap;

  va_start(ap, fmt);
  vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap);

  SysFatalAlert(buf);
}

void fatal_malloc(int size)
{
  /*fatal("OOM for size 0x%lx (%d)", size, size);*/
  debug(("OOM for size 0x%lx (%d)", size, size));
  ErrThrow(size);
}

#ifdef DEBUGGING
void debug_printf(char *fmt, ...)
{
  char buf[2048], *ptr, *start;
  va_list ap;

  va_start(ap, fmt);
  vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap);

  if (HostGetHostPlatform() == hostPlatformPalmOS) {
    DbgMessage(buf); DbgMessage("\n");
  }
  else {
    start = buf;
    while (*start) {
      for (ptr = start; *ptr != '\n' && *ptr != '\0'; ptr++) ;
      if (*ptr == '\0') {
        HostTraceOutputTL(appErrorClass, "%s", start);
        break;
      } else {
        *ptr = '\0';
        HostTraceOutputTL(appErrorClass, "%s", start);
        start = ptr+1;
      }
    }
  }
}
#endif

static void drect(const char *mess, RectangleType *rect) DRAW_SECT;
static void drect(const char *mess, RectangleType *rect)
{
  debug(("%s: (%d, %d) --> (%d, %d) [%dx%d]",
        mess, rect->topLeft.x, rect->topLeft.y,
        rect->topLeft.x + rect->extent.x,
        rect->topLeft.y + rect->extent.y,
        rect->extent.x, rect->extent.y));
}

/* --- midend --> frontend graphics --- */

/*
 * Since this front end does not support printing (yet), we need
 * this stub to satisfy the reference in midend_print_puzzle().
 */
void document_add_puzzle(document *doc, const game *game, game_params *par,
			 game_state *st, game_state *st2)
{
}

void frontend_default_colour(frontend *fe, float *output)
{
  RGBColorType col;
  UIColorGetTableEntryRGB(UIFormFill, &col);
  output[0] = col.r / 255.0;
  output[1] = col.g / 255.0;
  output[2] = col.b / 255.0;
}

void deactivate_timer(frontend *fe)
{
  debug(("deactivate_timer"));
  fe->tt = 0;
  fe->flags &= ~FF_TIMERON;
}

void activate_timer(frontend *fe)
{
  fe->flags |= FF_TIMERON;
  fe->tt = TimGetTicks();
  debug(("activate_timer tps=%u tt=%lu", tps, fe->tt));
}

void get_random_seed(void **randseed, int *randseedsize)
{
  long *v = snew(long);
  *v = TimGetSeconds();
  *randseed = v;
  *randseedsize = sizeof(long);
  debug(("get_random_seed gave seed 0x%lx", *v));
}

/* --- drawing functions --- */

#define sc_graphics 1
#define sc_text 2
static void set_colour(frontend *fe, int colour, RGBColorType *rgb, unsigned which) DRAW_SECT;
static void set_colour(frontend *fe, int colour, RGBColorType *rgb, unsigned which)
{
  /* Now, I could just use the idx for new as well as old models,
   * but I'm not convinced that new models that use 16/32-bit
   * graphics are using indexed bitmaps at all, and the docs
   * are a bit, er, hazy on what happens if you use the idx
   * functions with non-indexed bitmaps. */
  if (rgb_colour) {
    if (which & sc_graphics) WinSetForeColorRGB(rgb, NULL);
    if (which & sc_text) WinSetTextColorRGB(rgb, NULL);
  } else {
    IndexedColorType idx = (colour == -1) ? WinRGBToIndex(rgb) : fe->idxs[colour];
    if (which & sc_graphics) WinSetForeColor(idx);
    if (which & sc_text) WinSetTextColor(idx);
  }
}

static void set_colour_fe(frontend *fe, int colour, unsigned which) DRAW_SECT;
static void set_colour_fe(frontend *fe, int colour, unsigned which)
{
  int *ccol = (which == sc_graphics) ? &fe->gcol : &fe->tcol;
  if (colour != *ccol) {
    *ccol = colour;
    set_colour(fe, colour, &(fe->colours[colour]), which);
  }
}

static void palm_draw_text(void *handle, int x, int y, int fonttype, int fontsize,
               int align, int colour, char *text) DRAW_SECT;
static void palm_draw_text(void *handle, int x, int y, int fonttype, int fontsize,
               int align, int colour, char *text)
{
  frontend *fe = (frontend *)handle;
  int ptsz = SCALE(fontsize), i, swidth, fx = SCALE(x), fy = SCALE(y);
  fontinfo *fi = &fontinfos[0];
  FontID oldfont;
  WinDrawOperation prevdm;

  /* XXX All palm fonts are variable-pitch; if anything really
   * cares about fixed-pitch we should find a fixed-pitch data set
   * from under a rock somewhere. */

  /*debug(("text, %d,%d c%d size %d->%d '%s'", x, y, colour, fontsize,
        ptsz, text));*/

  for (i = 0; i < NFONTS; i++) {
    if (fontinfos[i].ysize > ptsz) goto bestfont;
    fi = &fontinfos[i];
  }
bestfont:
  debug(("best font match ID%d, size %d (requested was %d -> %d)",
         fi->fid, fi->ysize, fontsize, ptsz));
  oldfont = FntSetFont(fi->fid);
  swidth = FntCharsWidth(text, strlen(text));

  /*debug(("scaled pos %d, %d, %s %s", fx, fy,
        (align & ALIGN_VCENTRE) ? "VCENTRE" : "VNORMAL",
        (align & ALIGN_HCENTRE) ? "HCENTRE" :
        (align & ALIGN_HRIGHT) ? "HRIGHT" : "HLEFT"));*/
  if (align & ALIGN_VCENTRE)
    fy -= fi->ysize/2;
  else
    fy -= fi->ysize;
  if (align & ALIGN_HCENTRE)
    fx -= swidth / 2;
  else if (align & ALIGN_HRIGHT)
    fx -= swidth;

  set_colour_fe(fe, colour, sc_text);
  if (!rgb_colour && fe->idxs[colour] == 0) {
    IndexedColorType idxsave;
    debug (("palm_draw_text: indexed, white text"));
    /* In the case of indexed colour (older Palms like IIIC) the code below
     * (using winOverlay drawing mode and WinPaintChars()) does not work to
     * draw white characters (colour index 0 is white) on a coloured background.
     * Instead we set the background colour to the current pixel colour at fx,fy
     * and use WinDrawChars() to obtain the desired effect */
    idxsave = WinSetBackColor(WinGetPixel(fx, fy));
    WinDrawChars(text, strlen(text), fx, fy);
    WinSetBackColor(idxsave);
  }
  else if (!hd_features && fe->idxs[colour] == 0) {
    /* Special case for PalmOS 4.x (m505 and Clis with HRLib). We use the same approach
     * as above to get white characters on a coloured background. The only
     * difference is that we use the RGB versions of the functions instead
     * of the indexed colour versions. */
    RGBColorType rgb, rgbsave;
    debug(("palm_draw_text: clie, rgb, white text"));
    WinGetPixelRGB(fx, fy, &rgb);
    WinSetBackColorRGB(&rgb, &rgbsave);
    WinDrawChars(text, strlen(text), fx, fy);
    WinSetBackColorRGB(&rgbsave, NULL);
  }
  else {
    debug(("palm_draw_text: rgb overlay"));
    /* hurrah, this is how to display transparent fonts on PalmOS. */
    prevdm = WinSetDrawMode(winOverlay);
    WinPaintChars(text, strlen(text), fx, fy);
    WinSetDrawMode(prevdm);
  }
  FntSetFont(oldfont);
}

void palm_draw_rect(void *handle, int x, int y, int w, int h, int colour)
{
  frontend *fe = (frontend *)handle;
  //debug(("rect, %d,%d %dx%d c%d", x, y, w, h, colour));

  set_colour_fe(fe, colour, sc_graphics);

  if (w == 1 && h == 1) {
    WinDrawPixel(x, y);
  } else {
    RectangleType rect;
    rect.topLeft.x = SCALE(x);
    rect.topLeft.y = SCALE(y);
    rect.extent.x = SCALE(w);
    rect.extent.y = SCALE(h);
    WinDrawRectangle(&rect,0);
  }
}

static void palm_draw_line(void *handle, int x1, int y1, int x2, int y2, int colour) DRAW_SECT;
static void palm_draw_line(void *handle, int x1, int y1, int x2, int y2, int colour)
{
  frontend *fe = (frontend *)handle;

  //debug(("line, %d,%d to %d,%d c%d", x1, y1, x2, y2, colour));
  set_colour_fe(fe, colour, sc_graphics);
  WinDrawLine(SCALE(x1), SCALE(y1), SCALE(x2), SCALE(y2));
}

static void palm_draw_polygon(void *handle, int *coords, int npoints,
                  int fillcolour, int outlinecolour) DRAW_SECT;
static void palm_draw_polygon(void *handle, int *coords, int npoints,
                  int fillcolour, int outlinecolour)
{
  frontend *fe = (frontend *)handle;
  int x1, y1, x2, y2, i;

  if (fillcolour >= 0) {
    set_colour_fe(fe, fillcolour, sc_graphics);
    WinFillPolygon(coords, npoints, fe->sc, fe->sc);
  }

  assert(outlinecolour >= 0);
  set_colour_fe(fe, outlinecolour, sc_graphics);

  x1 = coords[0]; y1 = coords[1];
  for (i = 1; i < npoints; i++) {
    x2 = coords[i*2]; y2 = coords[i*2+1];
    palm_draw_line((void*)fe, x1, y1, x2, y2, outlinecolour);
    x1 = x2; y1 = y2;
  }
  x2 = coords[0]; y2 = coords[1];
  palm_draw_line((void*)fe, x1, y1, x2, y2, outlinecolour);
}

static void palm_clip(void *handle, int x, int y, int w, int h) DRAW_SECT;
static void palm_clip(void *handle, int x, int y, int w, int h)
{
  frontend *fe = (frontend *)handle;
  RectangleType rect;
  debug(("clip, %d,%d %dx%d", x, y, w, h));

  rect.topLeft.x = SCALE(x);
  rect.topLeft.y = SCALE(y);
  rect.extent.x = SCALE(w);
  rect.extent.y = SCALE(h);
  WinSetClip(&rect);
}

static void palm_unclip(void *handle) DRAW_SECT;
static void palm_unclip(void *handle)
{
  frontend *fe = (frontend *)handle;
  RectangleType rect;
  debug(("unclip"));

  rect.topLeft.x = 0;
  rect.topLeft.y = 0;
  rect.extent.x = fe->x;
  rect.extent.y = fe->y;

  WinSetClip(&rect);
}

static void palm_start_draw(void *handle) DRAW_SECT;
static void palm_start_draw(void *handle)
{
  frontend *fe = (frontend *)handle;

  debug(("start_draw"));
  assert(fe->bl && fe->bl->win_off);
  fe->old_draw = WinSetDrawWindow(fe->bl->win_off);
  WinPushDrawState();
  if (hd_features) WinSetCoordinateSystem(kCoordinatesNative);
  fe->gcol = fe->tcol = -1;
}

static void reset_update(frontend *fe) DRAW_SECT;
static void reset_update(frontend *fe)
{
  debug(("reset_update"));
  fe->x1up = fe->x + 1;
  fe->y1up = fe->y + 1;
  fe->x2up = fe->y2up = -1;
  fe->flags &= ~FF_UPDATEBMP;
}

static void palm_draw_update(void *handle, int x, int y, int w, int h) DRAW_SECT;
static void palm_draw_update(void *handle, int x, int y, int w, int h)
{
  frontend *fe = (frontend *)handle;

  /*debug(("draw_update (%d, %d) by %dx%d", x, y, w, h));*/
  fe->x1up = min(fe->x1up, SCALE(x));
  fe->y1up = min(fe->y1up, SCALE(y));
  fe->x2up = max(fe->x2up, SCALE(x+w));
  fe->y2up = max(fe->y2up, SCALE(y+h));
  fe->flags |= FF_UPDATEBMP;
  /*debug(("new update (%d, %d) --> (%d, %d)",
        fe->x1up, fe->y1up, fe->x2up, fe->y2up));*/
}

static void palm_end_draw(void *handle) DRAW_SECT;
static void palm_end_draw(void *handle)
{
  frontend *fe = (frontend *)handle;

  debug(("end_draw"));
  WinPopDrawState();
  WinSetDrawWindow(fe->old_draw);
  FrmUpdateForm(frmMain, 1);
}

static void palm_status_bar(void *handle, char *text) DRAW_SECT;
static void palm_status_bar(void *handle, char *text)
{
  frontend *fe = (frontend *)handle;

  char *rewritten = midend_rewrite_statusbar(fe_me(fe), text);
  debug(("status: '%s', rewritten: '%s'", text, rewritten));
  if (!fe->statusmsg || strcmp(rewritten, fe->statusmsg)) {
    sfree(fe->statusmsg);
    fe->statusmsg = rewritten;
    fe->flags |= FF_UPDATEST;
  } else {
    sfree(rewritten);
  }
}

/* --- blitter functions, and helper stuff --- */

static void palm_blitter_free(void *handle, blitter *bl) DRAW_SECT;
static void palm_blitter_free(void *handle, blitter *bl)
{
  if (!bl) return;
  if (bl->win_off) {
    WinDeleteWindow(bl->win_off, false);
    if (bl->bmp_off) BmpDelete(bl->bmp_off);
    if (bl->bmp_half) BmpDelete(bl->bmp_half);
  }
  sfree(bl);
}

static blitter *palm_blitter_new(void *handle, int w, int h) DRAW_SECT;
static blitter *palm_blitter_new(void *handle, int w, int h)
{
  blitter *bl = snew(blitter);
  memset(bl, 0, sizeof(blitter));
  bl->w = w;
  bl->h = h;
  return bl;
}

static void blitter_mkbitmap(frontend *fe, blitter *bl) DRAW_SECT;
static void blitter_mkbitmap(frontend *fe, blitter *bl)
{
  UInt16 err;
  UInt32 depth;
  ColorTableType *ct = NULL;
  int sw = SCALE(bl->w), sh = SCALE(bl->h);

  {
    UInt32 width, height;
    Boolean is_colour;

    // make sure depth is the same as our current screen depth.
    WinScreenMode(winScreenModeGet, &width, &height, &depth, &is_colour);
    debug(("current screen is %ldx%ld, %ld-bit, %s",
           width, height, depth, is_colour ? "colour" : "b+w"));
  }

#ifdef HAVE_CLIE_SDK
  if (clie) {
    debug(("HRBmpCreate %dx%d %ld-bit", sw, sh, depth));
    bl->bmp_off = HRBmpCreate(hrRefNum, sw, sh, depth, ct, &err);
  } else {
#endif
    debug(("BmpCreate: %dx%d %ld-bit", sw, sh, depth));
    bl->bmp_off = BmpCreate(sw, sh, (UInt8)depth, ct, &err);
#ifdef HAVE_CLIE_SDK
  }
#endif
  if (!bl->bmp_off) {
    debug(("(HR)BmpCreate failed, err = 0x%x", err));
    palm_blitter_free((void*)fe, bl);
    ErrThrow(err);
  }

  if (hd_features && scr_density > 1.0) {
    void *bits = BmpGetBits(bl->bmp_off);
    bl->bmp_half = bl->bmp_off;
    bl->bmp_off = (BitmapType *)BmpCreateBitmapV3(bl->bmp_half, kdensity,
                                                  bits, ct);
    if (!bl->bmp_off) {
      debug(("BmpCreateBitmapV3 failed."));
      palm_blitter_free((void*)fe, bl);
      ErrThrow(0);
    }
  } else
    bl->bmp_half = NULL;

  bl->win_off = WinCreateBitmapWindow(bl->bmp_off, &err);
  if (!bl->win_off) {
    debug(("WinCBW failed, err = 0x%x", err));
    palm_blitter_free((void*)fe, bl);
    ErrThrow(err);
  }

  debug(("Offscreen bitmap/window created."));
}

/* Another optimisation (which I originally tried) was to
 * create a 4-bit bitmap with a custom colour table,
 * to save even more space. However, it seems that the OS5
 * V3 bitmaps (which are needed for double-density blitting)
 * have nasty bugs to do with colour tables, that cause all
 * OS5 (but not OS6) simulators to crash. To replicate this,
 * set 'depth = fe->depth' and 'ct = fe->ct' at the start of
 * blitter_new, and watch them crash and burn. */

/* WinCopyRectangle(srcWin, destWin, srcRect, destX, destY, paintType) */

static void FixupCopyRectangle(WinHandle srcW, WinHandle dstW,
                               RectangleType *srcR,
                               Coord destX, Coord destY, WinDrawOperation mode) DRAW_SECT;
static void FixupCopyRectangle(WinHandle srcW, WinHandle dstW,
                               RectangleType *srcR,
                               Coord destX, Coord destY, WinDrawOperation mode)

{
#ifdef HAVE_CLIE_SDK
  if (clie)
    HRWinCopyRectangle(hrRefNum, srcW, dstW, srcR, destX, destY, mode);
  else
#endif
    WinCopyRectangle(srcW, dstW, srcR, destX, destY, mode);
}

static void palm_blitter_save(void *handle, blitter *bl, int x, int y) DRAW_SECT;
static void palm_blitter_save(void *handle, blitter *bl, int x, int y)
{
  frontend *fe = (frontend *)handle;
  RectangleType rect;

  if (!bl->win_off) blitter_mkbitmap(fe, bl);

  bl->x = x; bl->y = y;

  rect.topLeft.x = SCALE(bl->x);
  rect.topLeft.y = SCALE(bl->y);
  rect.extent.x = SCALE(bl->w);
  rect.extent.y = SCALE(bl->h);

  FixupCopyRectangle(fe->bl->win_off, bl->win_off,
                     &rect, 0, 0, winPaint);
}

static void palm_blitter_load(void *handle, blitter *bl, int x, int y) DRAW_SECT;
static void palm_blitter_load(void *handle, blitter *bl, int x, int y)
{
  frontend *fe = (frontend *)handle;

  if (x == BLITTER_FROMSAVED) x = bl->x;
  if (y == BLITTER_FROMSAVED) y = bl->y;

  assert(bl->win_off);

#ifdef HAVE_CLIE_SDK
  if (clie)
    HRWinDrawBitmap(hrRefNum, bl->bmp_off, SCALE(x), SCALE(y));
  else
#endif
    WinDrawBitmap(bl->bmp_off, SCALE(x), SCALE(y));
}

/* --- and the platform struct --- */

const struct drawing_api palm_drawing = {
    palm_draw_text,
    palm_draw_rect,
    palm_draw_line,
    palm_draw_polygon,
    palm_draw_circle,
    palm_draw_update,
    palm_clip,
    palm_unclip,
    palm_start_draw,
    palm_end_draw,
    palm_status_bar,
    palm_blitter_new,
    palm_blitter_free,
    palm_blitter_save,
    palm_blitter_load,
    NULL, NULL, NULL, NULL, NULL, NULL, /* {begin,end}_{doc,page,puzzle} */
    NULL,			       /* line_width */
};

/* --- other stuff --- */

static void DoPleaseWait(int clean) DRAW_SECT;
static void DoPleaseWait(int clean)
{
  Coord scrw, scrh;
  RectangleType r1, r2;

  WinGetDisplayExtent(&scrw, &scrh);
  r1.extent.x = 70;
  r1.topLeft.x = (scrw - r1.extent.x) / 2 + 1;
  r1.extent.y = 26;
  r1.topLeft.y = (scrh - r1.extent.y) / 2 - 5;
  r2.topLeft.x = r1.topLeft.x + 2;
  r2.topLeft.y = r1.topLeft.y + 2;
  r2.extent.x = 66;
  r2.extent.y = 22;
  if (clean) {
    WinEraseRectangle(&r1, 5);
    fe_global.flags &= ~FF_CLEANWAIT;
  } else {
    WinDrawRectangle(&r1, 5);
    WinEraseRectangle(&r2, 4);
    WinDrawChars("Please wait...", StrLen("Please wait..."),
                 r2.topLeft.x + 8, r2.topLeft.y + 5);
    fe_global.flags |= FF_CLEANWAIT;
  }
}

/*typedef struct pw_ctx {
  WinHandle oldActive;
  FormType *pWaitForm;
  FormActiveStateType oldState;
} pw_ctx;

static Boolean WaitHandleEvent(EventPtr event) DRAW_SECT;
static Boolean WaitHandleEvent(EventPtr event)
{
    Boolean handled = false;

    debug(("WaitHandleEvent"));
    switch (event->eType) {
    case frmLoadEvent:
      debug(("Wait had FrmLoadEvent"));
      break;

    case frmOpenEvent:
      debug(("Wait had FrmOpenEvent"));
      break;

    default:
    }
    return handled;
}


static void ShowPleaseWait(pw_ctx *ctx) DRAW_SECT;
static void ShowPleaseWait(pw_ctx *ctx)
{
  assert(ctx);

  debug(("ShowPleaseWait"));
  FrmSaveActiveState(&ctx->oldState);
  ctx->oldActive = WinGetDrawWindow();

  ctx->pWaitForm = FrmInitForm(frmWait);
  FrmSetEventHandler(ctx->pWaitForm, WaitHandleEvent);
  FrmSetActiveForm(ctx->pWaitForm);
  FrmDrawForm(ctx->pWaitForm);
}

static void HidePleaseWait(pw_ctx *ctx) DRAW_SECT;
static void HidePleaseWait(pw_ctx *ctx)
{
  assert(ctx);

  debug(("HidePleaseWait"));
  FrmEraseForm(ctx->pWaitForm);
  FrmDeleteForm(ctx->pWaitForm);
  WinSetDrawWindow(ctx->oldActive);
  FrmRestoreActiveState(&ctx->oldState);
}*/

static void DrawPuzzle(frontend *fe, UInt16 code) DRAW_SECT;
static void DrawPuzzle(frontend *fe, UInt16 code)
{
  RectangleType rect, strect;
  RGBColorType rgb;
  int updatebmp = 1, updatest = (fe->statush > 0) ? 1 : 0;

  if (fe->flags & FF_CLEANWAIT) {
    DoPleaseWait(1);
    code = frmRedrawUpdateCode;
  }

  WinPushDrawState();
  if (hd_features) WinSetCoordinateSystem(kCoordinatesNative);

  switch (code) {
  case frmRedrawUpdateCode:
    WinGetDrawWindowBounds(&rect);
    debug(("DrawPuzzle whole puzzle:"));
    drect("draw window", &rect);
    UIColorGetTableEntryRGB(UIFormFill, &rgb);
    set_colour(fe, -1, &rgb, sc_graphics);
    WinDrawRectangle(&rect, 0);

    if (fe->sc == 1.0) {
      /* screen pixels, for the outline rectangle */
      rect.topLeft.x = fe->xoff;
      rect.topLeft.y = fe->yoff;
      rect.extent.x = fe->x;
      rect.extent.y = fe->y;
      rgb.r = rgb.g = rgb.b = 0;
      drect("board frame", &rect);
      WinDrawRectangleFrame(rectangleFrame, &rect);
    }

    /* bitmap pixels, for the blit operation */
    rect.topLeft.x = 0;
    rect.topLeft.y = 0;
    rect.extent.x = fe->x;
    rect.extent.y = fe->y;
    break;

  case 1:
    debug(("DrawPuzzle update region."));
    if (!(fe->flags & FF_UPDATEST))
      updatest = 0;
    if (!(fe->flags & FF_UPDATEBMP))
    {
      updatebmp = 0;
    } else {
      rect.topLeft.x = fe->x1up;
      rect.topLeft.y = fe->y1up;
      rect.extent.x = fe->x2up - fe->x1up;
      rect.extent.y = fe->y2up - fe->y1up;
    }
    break;

  default:
    debug(("DrawPuzzle unknown code %d.", code));
    break;
  }

  if (updatest) {
    FontID oldfont;
    int statush = fe->statush;
    if (clie) statush /= scr_density;

    debug(("updating status msg."));
    /* remove old msg in case new one is shorter. */
    strect.topLeft.x = 0;
    strect.topLeft.y = 0;
    strect.extent.x = fe->wscr * scr_density;
    strect.extent.y = statush;
    drect("erase status", &strect);
    WinEraseRectangle(&strect, 0);

    UIColorGetTableEntryRGB(UIObjectForeground, &rgb);
    set_colour(fe, -1, &rgb, sc_graphics | sc_text);

    WinDrawLine(0, statush-1, fe->wscr * scr_density, statush-1);
    if (fe->statusmsg) {
      oldfont = FntSetFont(stdFont);
      WinDrawChars(fe->statusmsg, strlen(fe->statusmsg),
                   2, 0);
      FntSetFont(oldfont);
    }
    fe->flags &= ~FF_UPDATEST;
  }

  if (updatebmp) {
    drect("about to copy", &rect);
    FixupCopyRectangle(fe->bl->win_off,
                       WinGetDrawWindow(),
                       &rect,
                       fe->xoff + rect.topLeft.x,
                       fe->yoff + rect.topLeft.y,
                       winPaint);
  }
  WinPopDrawState();
  reset_update(fe);
}

static void fettle_screen(frontend *fe) PUZZ_SECT;
static void fettle_screen(frontend *fe)
{
  UInt32 depth, alldepth, i, maxdepth;
  Boolean colour = true;

  rgb_colour = 0;
  WinScreenMode(winScreenModeGetSupportedDepths,
                NULL, NULL, &alldepth, &colour /* in */);
  WinScreenMode(winScreenModeGet,
                &fe->wscr, &fe->hscr, &depth, &colour);
  debug(("screen details: %lux%lu, %lu-bit %s",
        fe->wscr, fe->hscr, depth, colour ? "colour" : "b+w"));
  for (i = 0; i < 32; i++) {
    if (alldepth & (2<<i)) {
      debug(("supports %lu-bit graphics", i+2));
      maxdepth = i+2;
    }
  }
  if ((depth < 16) && (maxdepth > depth)) {
    debug(("Attempting to raise current %ld-bit screen to %ld bits.",
           depth, maxdepth));
    WinScreenMode(winScreenModeSet, NULL, NULL, &maxdepth, NULL);
  }
  WinScreenMode(winScreenModeGet, NULL, NULL, &depth, NULL);
  if (depth >= 16) {
    rgb_colour = 1;
    debug(("Direct rgb colour supported"));
  }
}

/* --- frontend new game creation --- */

static void clearfe_game(frontend *fe) PUZZ_SECT;
static void clearfe_game(frontend *fe)
{
  if (!fe) return;
  palm_blitter_free((void*)fe, fe->bl);
  fe->bl = NULL;
  if (fe->ct) {
    sfree(fe->ct);
    fe->ct = NULL;
  }
  if (fe->idxs) {
    sfree(fe->idxs);
    fe->idxs = NULL;
  }
  if (fe->statusmsg) {
    sfree(fe->statusmsg);
    fe->statusmsg = NULL;
  }
  reset_update(fe); deactivate_timer(fe);
}

static void clearfe(frontend *fe) PUZZ_SECT;
static void clearfe(frontend *fe)
{
  int i;

  if (!fe) return;
  clearfe_game(fe);
  if (fe->pergame) {
    for (i = 0; i < gamecount; i++)
      midend_free(fe->pergame[i].me);
    sfree(fe->pergame);
  }
}

static void app_new_size(frontend *fe, RectangleType *drawsize, int newgame) PUZZ_SECT;
static void app_new_size(frontend *fe, RectangleType *drawsize, int newgame)
{
  midend *me = fe_me(fe);
  int x, y, ox, oy;
  float wsc, hsc;

  /* --- here, screen size is in Palm coordinates. */
  if (fe->wscr == drawsize->extent.x &&
      fe->hscr == drawsize->extent.y &&
      !newgame) return;
  fe->wscr = drawsize->extent.x;
  fe->hscr = drawsize->extent.y;

  drect("drawsize (screen coords)", drawsize);
  drawsize->extent.x *= scr_density;
  drawsize->extent.y *= scr_density;

  /* --- from here on, screen size is in pixels. */
  drect("drawsize (pixels)", drawsize);

  if (midend_wants_statusbar(me)) {
    FontID oldfont;
    int statussize;

    oldfont = FntSetFont(stdFont); /* configure font? */
    statussize = FntCharHeight() + 1;
    fe->statush = statussize * scr_density;
    drawsize->extent.y -= fe->statush;
    FntSetFont(oldfont);
    debug(("status bar height %d", fe->statush));
  } else {
    fe->statush = 0;
    debug(("No status bar."));
  }

  x = drawsize->extent.x;
  y = drawsize->extent.y;
  midend_size(me, &x, &y, FALSE);
  debug(("New game size %dx%d", x, y));

  wsc = ((float)drawsize->extent.x)/((float)x);
  hsc = ((float)drawsize->extent.y)/((float)y);
  fe->sc = min(wsc, hsc);
  if (fe->sc > 4.0) fe->sc = 4.0;
  else if (fe->sc > 2.0) fe->sc = 2.0;
  else if (fe->sc > 1.0) fe->sc = 1.0;
  debug(("wsc = %d%%, hsc = %d%%, sc = %f%%",
        (int)(wsc*100), (int)(hsc*100), fe->sc*100.0));

  ox = fe->x; oy = fe->y;
  fe->x = SCALE(x); fe->y = SCALE(y);
  if (fe->x != ox || fe->y != oy || newgame) {

    if (fe->bl) palm_blitter_free((void*)fe, fe->bl);
    fe->bl = palm_blitter_new((void*)fe, x, y); /* must pass unscaled */
    blitter_mkbitmap(fe, fe->bl);
    if (!newgame) midend_redraw(me);
  }

  if (fe->x >= drawsize->extent.x)
    fe->xoff = 0;
  else
    fe->xoff = ((drawsize->extent.x) - fe->x) / 2;
  if (fe->y >= drawsize->extent.y)
    fe->yoff = 0;
  else
    fe->yoff = ((drawsize->extent.y) - fe->y) / 2;
  fe->yoff += fe->statush;
  debug(("xoff=%d, yoff=%d", fe->xoff, fe->yoff));
}

/* returns non-zero on error (or cancel from something) */
static int app_new_game(frontend *fe) PUZZ_SECT;
static int app_new_game(frontend *fe)
{
  int i;
  float *colours;
  RectangleType drawsize;
  midend *me = fe_me(fe);

  WinGetDrawWindowBounds(&drawsize);

  DoPleaseWait(0);

  midend_stop_anim(fe_me(fe));
  clearfe_game(fe);

  if (fe->flags & FF_NONEWGAME) {
    debug(("Skipping new game generation, serialised has been loaded"));
    fe->flags &= ~FF_NONEWGAME;
  } else {
    UInt32 tt_now = TimGetTicks();
    float elapsed;
    UInt8 brightness = SysLCDBrightness(true, 0);
    midend_new_game(me);
    SysLCDBrightness(true, brightness);
    elapsed =  (TimGetTicks() - tt_now) / (float)tps;
    debug(("Game generation for %s took %g seconds.",
           gamelist[fe->currgame]->name, (double)elapsed));
  }

  /* Set up colour table for new game. */
  colours = midend_colours(me, &fe->ncolours);
  if (fe->ncolours > 256) fe->depth = 16;
  else if (fe->ncolours > 16) fe->depth = 8;
  else fe->depth = 4;
  fe->ct = smalloc(sizeof(ColorTableType) + ((1<<fe->depth) * sizeof(RGBColorType)));
  fe->ct->numEntries = 1<<fe->depth;
  fe->colours = ColorTableEntries(fe->ct);
  fe->idxs = snewn(fe->ct->numEntries, IndexedColorType);
  for (i = 0; i < fe->ncolours; i++) {
    fe->colours[i].index = 0;
    fe->colours[i].r = (UInt8)(colours[i*3+0]*255.0);
    fe->colours[i].g = (UInt8)(colours[i*3+1]*255.0);
    fe->colours[i].b = (UInt8)(colours[i*3+2]*255.0);
    debug(("Colour %d r%d g%d b%d",
          i, fe->colours[i].r, fe->colours[i].g, fe->colours[i].b));
    fe->idxs[i] = WinRGBToIndex(&fe->colours[i]);
  }
  for (; i < fe->ct->numEntries; i++) {
    memset(&(fe->colours[i]), 0, sizeof(RGBColorType));
    /*debug(("Colour %d blank", i));*/
  }
  sfree(colours);

  app_new_size(fe, &drawsize, 1);

  reset_update(fe);
  deactivate_timer(fe);

  midend_redraw(me);
  fe->flags |= FF_WANTSTORE;

  return 0;
}

int app_change_game(frontend *fe, int gameno) PUZZ_SECT;
int app_change_game(frontend *fe, int gameno)
{
  FormType *frmP;

  debug(("app_change_game, gameno %d", gameno));
  assert(gameno <= gamecount);

  if ((gameno != fe->currgame) && (fe->flags & FF_GAMELOCKED)) {
    debug(("app_change_game: attempt to change game while locked."));
    FrmAlert(gameLocked);
    return -1;
  }

  midend_stop_anim(fe_me(fe)); /* otherwise sixteen.c assert()s. */

  fe->currgame = gameno;

  app_new_game(fe);

  frmP = FrmGetFormPtr(frmMain);
  if (frmP) {
    debug(("Reloading menu for new presets for new game."));
    FrmSetMenu(frmP, menuMain);
  }
  return 0;
}

/* --- Reading our font list --- */
/*
 * PalmOS doesn't have a 'give me a font of size X' call; it all runs
 * on fixed-size bitmap fonts. Therefore, we read in a variety of
 * differently-sized fonts at startup, order them by size, and pick
 * the best of these based on the size we're asked for (in draw_text).
 */

static void ZeroFont(fontinfo *fi) DRAW_SECT;
static void ZeroFont(fontinfo *fi)
{
  fi->fid = 0;
  fi->fonth = NULL;
  fi->fontp = NULL;
  fi->ysize = 9999; /* will never be chosen. */
}

static void InitFont(int off, FontID fid, Boolean needload, Boolean v2) DRAW_SECT;
static void InitFont(int off, FontID fid, Boolean needload, Boolean v2)
{
  fontinfo *fi = &(fontinfos[off]);
  FontID ftmp;

  debug(("font #%d id %d", off, fid));
  fi->fid = fid;
  if (needload) {
    if (v2) {
      debug(("loading v2 font"));
      fi->fonth = DmGetResource(fontExtRscType, fid);
    } else {
      debug(("loading v1 font"));
      fi->fonth = DmGetResource(fontRscType, fid);
    }
    if (!fi->fonth) {
      debug(("unable to load font ID #%d", fid));
      ZeroFont(fi);
      return;
    }
    fi->fontp = (FontType *)MemHandleLock(fi->fonth);
    if (!fi->fontp) fatal("unable to lock font");
    FntDefineFont(fid, fi->fontp);
  } else {
    fi->fonth = NULL;
    fi->fontp = NULL;
  }

  WinPushDrawState();
  if (hd_features) WinSetCoordinateSystem(kCoordinatesNative);

  ftmp = FntSetFont(fid);
  fi->ysize = FntCharHeight();
  debug(("ysize %d", fi->ysize));
  FntSetFont(ftmp);

  WinPopDrawState();
}

static int comp_fsize(const void *v1, const void *v2) DRAW_SECT;
static int comp_fsize(const void *v1, const void *v2)
{
  const fontinfo *f1 = (const fontinfo *)v1;
  const fontinfo *f2 = (const fontinfo *)v2;

  if (f1->ysize < f2->ysize)
    return -1;
  else if (f1->ysize > f2->ysize)
    return 1;
  else return 0;
}

static void InitAllFonts(void) DRAW_SECT;
static void InitAllFonts(void)
{
  int i;

  /* we're now using either all hires or all lores. */
  for (i = 0; i < NFONTS_LORES; i++) {
    if (hd_features && scr_density > 1.0)
      ZeroFont(&(fontinfos[i]));
    else
      InitFont(i, 128+i, true, false);
  }
  for (/* leave */;i < NFONTS_LORES + NFONTS_HIRES; i++) {
    if (hd_features && scr_density > 1.0)
      InitFont(i, 128+i, true, false);
    else {
      ZeroFont(&(fontinfos[i]));
    }
  }
  InitFont(i++, stdFont, false, false);
  InitFont(i++, largeFont, false, false);

  debug(("sorting fonts..."));
  qsort(&fontinfos, NFONTS, sizeof(fontinfo), comp_fsize);
  for (i = 0; i < NFONTS; i++)
    debug(("font #%d size %d", fontinfos[i].fid, fontinfos[i].ysize));
}

static void CloseFonts(void) DRAW_SECT;
static void CloseFonts(void)
{
  int i;
  for (i = 0; i < NFONTS; i++) {
    if (fontinfos[i].fontp) {
      MemHandleUnlock(fontinfos[i].fonth);
      DmReleaseResource(fontinfos[i].fonth);
    }
  }
}

/* --- Start/Stop Application function --- */

static UInt16 StartApplication(void) PUZZ_SECT;
static UInt16 StartApplication(void)
{
  frontend *fe = &fe_global;
  int i;

  if (dia_features) {
    UInt16 card;
    LocalID db;
    Err err;
    /* Register for display resize notification */
    err = SysCurAppDatabase(&card, &db);
    if (err != errNone) return 1;
    err = SysNotifyRegister(card, db, sysNotifyDisplayResizedEvent, NULL,
            sysNotifyNormalPriority, 0);
    if (err != errNone) return 1;
  }

  memset(fe, 0, sizeof(frontend));

  fettle_screen(fe);
  InitAllFonts();

  fe->pergame = snewn(gamecount, struct pergame_data);
  for (i = 0; i < gamecount; i++) {
    fe->pergame[i].me = midend_new(fe, gamelist[i], &palm_drawing, (void*)fe);
  }

  ReadPrefs();
  if (app_change_game(fe, fe->currgame)) {
    /* couldn't create the saved game for some reason.
     * Try again with an easy-to-create game. */
    fe->currgame = 0;
    if (app_change_game(fe, fe->currgame))
      return 1;
  }
  FrmGotoForm(frmMain);
  return 0;
}

static void StopApplication(int ret) PUZZ_SECT;
static void StopApplication(int ret)
{
  FrmCloseAllForms();
  if (hd_features)
    WinScreenMode(winScreenModeSetToDefaults,
                  NULL, NULL, NULL, NULL);
#ifdef HAVE_CLIE_SDK
  if (clie) {
    HRWinScreenMode(hrRefNum, winScreenModeSetToDefaults,
                     NULL, NULL, NULL, NULL );
    HRClose(hrRefNum);
  }
#endif
  CloseFonts();
  if (!ret) {
    /* We only write prefs and clean up after ourselves if !ret
     * (i.e. we didn't throw the OOM exception). If we tried to
     * clear up we might end up attempting to free uninitialised
     * memory; we assume the OS can clear up after us anyway,
     * and we'd rather not cause segfaults. We also don't want
     * to save a potentially-dangerous preset to try and load
     * (and fail) again immediately we restart. */
    WritePrefs();
    clearfe(&fe_global); /* keep after WritePrefs! */
  }

  if (dia_features) {
    UInt16 card;
    LocalID db;
    Err err;
    err = SysCurAppDatabase(&card, &db);
    if (err == errNone) {
      SysNotifyUnregister(card, db, sysNotifyDisplayResizedEvent,
           sysNotifyNormalPriority);
    }
  }
}

/* --- 'Main form' (play area) event handling --- */

/* These next two might seem scungy (always having the menu item,
 * and then complaining at launch time), but that's actually what
 * the Palm UI guidelines say to do. */

static Boolean MainFormCopy(frontend *fe) PUZZ_SECT;
static Boolean MainFormCopy(frontend *fe)
{
  if (!midend_can_format_as_text_now(fe_me(fe))) {
    FrmCustomAlert(puzUnsupported,
                   gamelist[fe->currgame]->name,
                   "saving as text",
                   NULL);
  } else {
    char *text = midend_text_format(fe_me(fe));
    if (text) {
      ClipboardAddItem(clipboardText,
                       text, strlen(text));
      sfree(text);
    }
  }
  return true;
}

static Boolean MainFormSolve(frontend *fe) PUZZ_SECT;
static Boolean MainFormSolve(frontend *fe)
{
  if (!gamelist[fe->currgame]->can_solve) {
    FrmCustomAlert(puzUnsupported,
                   gamelist[fe->currgame]->name,
                   "automatic solving",
                   NULL);
  } else {
    char *msg = midend_solve(fe_me(fe));
    if (msg) FrmCustomAlert(puzSolveFailed, msg, NULL, NULL);
  }
  return true;
}


static Boolean MainFormDoKeystroke(frontend *fe, UInt16 key, int mod) PUZZ_SECT;
static Boolean MainFormDoKeystroke(frontend *fe, UInt16 key, int mod)
{
  int realkey = 0;
  static int n_ls = 0;

  debug(("MainFormDoKeystroke: key = '%c', %d (0x%x), mod = 0x%x",
         (char)key, key, key, mod));

  /* sekrit debugging mode, so we can gremlin one game at a time. */
  if (key == 'l') {
    n_ls++;
    if (n_ls == 10) {
      if (fe->flags & FF_GAMELOCKED)
        fe->flags &= ~FF_GAMELOCKED;
      else
        fe->flags |= FF_GAMELOCKED;
      debug(("you found the super-sekrit code! game is now %slocked",
             (fe->flags & FF_GAMELOCKED) ? "" : "un"));
      n_ls = 0;
    }
  } else
    n_ls = 0;

  switch (key) {
  case chrLeftArrow:            case vchrRockerLeft:
  case vchrThumbWheelUp:
#ifdef HAVE_CLIE_SDK
  case vchrJogUp:
#endif
    realkey = CURSOR_LEFT; break;

  case chrRightArrow:           case vchrRockerRight:
  case vchrThumbWheelDown:
#ifdef HAVE_CLIE_SDK
  case vchrJogDown:
#endif
    realkey = CURSOR_RIGHT; break;

  case chrUpArrow:              case vchrRockerUp:
  case vchrPageUp:
    realkey = CURSOR_UP; break;

  case chrDownArrow:            case vchrRockerDown:
  case vchrPageDown:
    realkey = CURSOR_DOWN;  break;

  case chrCarriageReturn:       case vchrRockerCenter:
  case vchrHardRockerCenter:    case vchrThumbWheelPush:
#ifdef HAVE_CLIE_SDK
  case vchrJogPush:
#endif
    realkey = CURSOR_SELECT; break;

  case 'c': case 'C': return MainFormCopy(fe);
  case 's': case 'S': return MainFormSolve(fe);

  case 'q': case 'Q': break; /* don't want quit msgs. */

  case 'n': case 'N':
    app_new_game(fe);
    break;

  default:
    realkey = key;
    break;
  }
  if (realkey) {
    debug(("realkey = '%c', %d (0x%x)",
          (char)realkey, realkey, realkey));
    if (!midend_process_key(fe_me(fe),
                            0, 0, realkey | mod)) {
      debug(("midend_process_key signalled to quit."));
      ErrThrow(1);
    }
    return true;
  } else
    return false;
}

/* Empirically, the max. no. of lines we can fit in a menu is 13. */
#define MAX_MENU_LINES 13

static Boolean MainFormInitMenu(frontend *fe) PUZZ_SECT;
static Boolean MainFormInitMenu(frontend *fe)
{
  int i, max_presets, n_presets;
  UInt16 nextid;

  debug(("MainFormInitMenu"));

  max_presets = MAX_MENU_LINES - 1; /* always a 'Default' line. */
  if (gamelist[fe->currgame]->can_configure) max_presets--;

  nextid = PRESET_MENU_START;
  n_presets = midend_num_presets(fe_me(fe));

  /* Arbitrarily take the first N presets, if we're presented with too many. */
  if (n_presets > max_presets) n_presets = max_presets;

  for (i = 0; i < n_presets; i++, nextid++) {
    game_params *ptmp;
    char *ntmp;
    midend_fetch_preset(fe_me(fe), i, &ntmp, &ptmp);
    debug(("Adding preset menu item %ud '%s'", nextid, ntmp));
    MenuAddItem(nextid-1, nextid, 0, ntmp);
  }
  if (gamelist[fe->currgame]->can_configure) {
    debug(("Adding preset menu item %ud 'Custom...'", PRESET_MENU_CUSTOM));
    MenuAddItem(nextid-1, PRESET_MENU_CUSTOM, 0, "Custom...");
  }
  return true;
}

static Boolean MainFormDoMenu(frontend *fe, UInt16 command) PUZZ_SECT;
static Boolean MainFormDoMenu(frontend *fe, UInt16 command)
{
  Boolean handled = false;
  int restart = 0;

  MenuEraseStatus(0);
  debug(("MainFormDoMenu: command %d", command));

  switch (command)
  {
  case gameChoose:
    FrmGotoForm(frmSelect);
    handled = true;
    break;

  case gameNew:
    restart = 1;
    break;

  case gameRestart:
    midend_restart_game(fe_me(fe));
    break;

  case gameSpecific:
  case gameSeed:
    /* might not change game, if cancelled. */
    if (!get_config(fe, (command == gameSpecific) ? CFG_DESC : CFG_SEED))
      restart = 1;
    handled = true;
    break;

  case gameUndo:    return MainFormDoKeystroke(fe, 'u', 0);
  case gameRedo:    return MainFormDoKeystroke(fe, '\x12', 0);

  case gameCopy:    return MainFormCopy(fe);
  case gameSolve:   return MainFormSolve(fe);

  case optionsRightClick:
    get_config(fe, CFG_PALM_RCLICK);
    handled = true;
    break;

  case helpAbout:
    {
      FormType *frmP = FrmInitForm(frmAbout);
      int labellen = strlen(FrmGetLabel(frmP, aboutVersion));
      debug(("current version string is %s (length %d)",
             FrmGetLabel(frmP, aboutVersion), labellen));
      if (labellen >= strlen(ver))
        FrmCopyLabel(frmP, aboutVersion, ver);
      else
        FrmCopyLabel(frmP, aboutVersion, "...too long...");
#ifdef HAVE_HANDERA_SDK
      if (handera) VgaFormModify(frmP, vgaFormModify160To240);
#endif
      FrmDoDialog(frmP);
      FrmDeleteForm(frmP);
      handled = true;
    }
    break;

  case helpHelp:
    FrmHelp(hlpHelp0 + fe->currgame);
    handled = true;
    break;

  case helpCommon:
    FrmHelp(hlpHelpCommon);
    handled = true;
    break;

  case PRESET_MENU_DEFAULT:
    {
      game_params *params = gamelist[fe->currgame]->default_params();
      midend_set_params(fe_me(fe), params);
      sfree(params);
      restart = 1;
      handled = true;
    }
    break;

  case PRESET_MENU_CUSTOM:
    if (!get_config(fe, CFG_SETTINGS))
      restart = 1;
    handled = true;
    break;

  default:
    break;
  }

  if (command >= PRESET_MENU_START && command <= PRESET_MENU_MAX) {
    game_params *ptmp;
    char *ntmp;
    int presetno = command - PRESET_MENU_START;

    restart = 1;
    assert(presetno < midend_num_presets(fe_me(fe)));
    midend_fetch_preset(fe_me(fe), presetno, &ntmp, &ptmp);
    midend_set_params(fe_me(fe), ptmp);
  }
  if (restart) {
    if (app_new_game(fe)) {
      FrmSetMenu(FrmGetFormPtr(frmMain), menuMain);
    }
  }

  return handled;

}

static Boolean MainFormDoPen(frontend *fe, int lchar, int rchar,
                             UInt16 x, UInt16 y)
{
  int gx, gy; /* to gane coords. */
  int realchar = (fe->flags & FF_ISRIGHT) ? rchar : lchar;

  x *= scr_density; y *= scr_density;
  debug(("MainFormDoPen: %s (%d), (%d, %d)",
         fe->flags & FF_ISRIGHT ? "right" : "left", realchar, x, y));

  gx = RSCALE(x - fe->xoff);
  gy = RSCALE(y - fe->yoff);
  debug(("MainFormDoPen: to midend (%d, %d)", gx, gy));

  if ((gx >= 0) && (gx < RSCALE(fe->x)) &&
      (gy >= 0) && (gy < RSCALE(fe->y))) {
    debug(("MainFormDoPen: OK to send to midend."));
    if (!midend_process_key(fe_me(fe),
                            gx, gy, realchar)) {
      debug(("midend_process_key signalled to quit."));
      ErrThrow(1);
    }
    return true;
  }
  return false;
}

#define ISRIGHT_TICKS  ((UInt32)((tps*options->rclick_millis)/1000L))
/*#define ISRIGHT(fe) ((gamelist[(fe)->currgame].flags & REQUIRE_RBUTTON) && \
                     ( (TimGetTicks() - (fe)->down_tt) > ISRIGHT_TICKS) )*/
#define ISRIGHT(fe) ((TimGetTicks() - (fe)->down_tt) > ISRIGHT_TICKS)

static void MainFormPenDown(frontend *fe) PUZZ_SECT;
static void MainFormPenDown(frontend *fe)
{
  debug(("sending pen down, clearing DOWNPENDING"));
  MainFormDoPen(fe, LEFT_BUTTON, RIGHT_BUTTON, fe->down_x, fe->down_y);
  fe->flags &= ~FF_DOWNPENDING;
}

static int CanIgnoreMove(UInt16 x1, UInt16 y1, UInt16 x2, UInt16 y2) PUZZ_SECT;
static int CanIgnoreMove(UInt16 x1, UInt16 y1, UInt16 x2, UInt16 y2)
{
  UInt16 diffsq = ((x1-x2) * (x1-x2)) + ((y1-y2) * (y1-y2));
  UInt16 noofpixels = options->rclick_offset;

  if (diffsq < noofpixels*noofpixels) return 1;
  return 0;
}

void SetupDIA(FormPtr frmP)
{
  /* Set up some magic for palms with DIA. */
  if (dia_features) {
    WinHandle frmWH = FrmGetWindowHandle(frmP);
    /* Slightly arbitrary constraints because we can size to whatever we
     * like. NB that width /can/ change (contrary to the docs) if the user
     * switches to landscape mode; that's all done by the OS, and all the app
     * sees is your 160x240 screen changing to a 240x160 screen. */
    WinSetConstraintsSize(frmWH,
                          80, 160, 512, /* min/pref/max height */
                          80, 160, 512);/* min/pref/max width */
    FrmSetDIAPolicyAttr(frmP, frmDIAPolicyCustom);
    PINSetInputTriggerState(pinInputTriggerEnabled);
    PINSetInputAreaState(pinInputAreaUser);
  }
}

static Boolean MainFormHandleEvent(EventPtr event) PUZZ_SECT;
static Boolean MainFormHandleEvent(EventPtr event)
{
  FormPtr frm;
  Boolean handled = false;
  frontend *fe = &fe_global;

  static UInt16 update_deferred = 0;
  static Boolean entered = 0;

  switch (event->eType) {
  case penDownEvent:
    if (event->screenY * scr_density < fe->statush) {
      debug(("Queueing up false menu click."));
      EvtEnqueueKey(vchrMenu, 0, commandKeyMask);
      return true;
    }

    debug(("penDown: tapcount = %d, setting DOWNPENDING", event->tapCount));
    fe->flags |= FF_DOWNPENDING;
    fe->flags &= ~FF_ISRIGHT;
    fe->down_tt = TimGetTicks();
    fe->down_x = event->screenX;
    fe->down_y = event->screenY;
    debug(("down_tt = %lu, location = (%d, %d)",
           fe->down_tt, fe->down_x, fe->down_y));
    handled = true;
    break;

  case penMoveEvent:
    if (event->screenY * scr_density < fe->statush) return true;
    if (fe->flags & FF_DOWNPENDING) {
      debug(("penMove but DOWNPENDING"));
      /* If we're waiting for a down (haven't decided whether or
       * not we're a left or right button), ignore small pen movements. */
      if (CanIgnoreMove(event->screenX, event->screenY,
                        fe->down_x, fe->down_y)) {
        debug(("penMove decides can ignore"));
        handled = true;
        break;
      }
      if (ISRIGHT(fe)) {
        debug(("penMove decides ISRIGHT"));
        fe->flags |= FF_ISRIGHT;
      }
      MainFormPenDown(fe);
    }
    debug(("penMove: sending DRAG"));
    return MainFormDoPen(fe, LEFT_DRAG, RIGHT_DRAG,
               event->screenX,
               event->screenY);

  case penUpEvent:
    if (event->screenY * scr_density < fe->statush) return true;
    if (fe->flags & FF_DOWNPENDING) {
      debug(("penUp but DOWNPENDING"));
      if (ISRIGHT(fe)) {
        debug(("penUp decides ISRIGHT"));
        fe->flags |= FF_ISRIGHT;
      }
      MainFormPenDown(fe);
    }
    debug(("penMove: sending RELEASE"));
    return MainFormDoPen(fe, LEFT_RELEASE, RIGHT_RELEASE,
               event->screenX,
               event->screenY);

  case menuOpenEvent:
    return MainFormInitMenu(&fe_global);

  case menuEvent:
    return MainFormDoMenu(&fe_global, event->data.menu.itemID);

  case keyDownEvent:
    {
      UInt16 chr = event->data.keyDown.chr;
#ifdef DEBUGGING
      UInt16 keycode = event->data.keyDown.keyCode;
#endif
      int mod = 0;
      Direction dir;

      debug(("keyDownEvent: chr = 0x%x ('%c', %d), keycode = 0x%x (%d), modifiers = 0x%x",
             chr, chr, chr, keycode, keycode, event->data.keyDown.modifiers));

      if (fiveway_key(event, &dir)) {
        switch (dir) {
        case Up:
          chr = CURSOR_UP;
          break;
        case Down:
          chr = CURSOR_DOWN;
          break;
        case Left:
          chr = CURSOR_LEFT;
          break;
        case Right:
          chr = CURSOR_RIGHT;
          break;
        case Center:
          chr = CURSOR_SELECT;
          break;
        default:
          return false;
        }
      }

      if (event->data.keyDown.modifiers & shiftKeyMask)
        mod |= MOD_SHFT;
      if (event->data.keyDown.modifiers & controlKeyMask)
        mod |= MOD_CTRL;

      return MainFormDoKeystroke(&fe_global, chr, mod);
    }

  case frmUpdateEvent:
    if (!entered) {
      debug(("deferring update request"));
      update_deferred = event->data.frmUpdate.updateCode;
    } else
      DrawPuzzle(fe, event->data.frmUpdate.updateCode);
    handled = true;
    break;

  case frmOpenEvent:
    frm = FrmGetFormPtr(event->data.frmOpen.formID);
#ifdef HAVE_HANDERA_SDK
    if (handera) VgaFormModify(frm, vgaFormModify160To240);
#endif
    debug(("frmOpen: formID %d, form 0x%x",
          event->data.frmOpen.formID, frm));
    FrmDrawForm(frm);

    {
      /* because sodding UX-50s open their DIA without sending an event */
      RectangleType newbounds;
      FixupWinGetBounds(WinGetDisplayWindow(), &newbounds);
      app_new_size(fe, &newbounds, 0);
    }

    DrawPuzzle(fe, frmRedrawUpdateCode);
    handled = true;
    break;

  case winExitEvent:
    if (event->data.winExit.exitWindow ==
        (WinHandle)FrmGetFormPtr(frmMain)) {
      debug(("winExitEvent for main window"));
      entered = 0;
    }
    break;

  case winEnterEvent:
    if (event->data.winEnter.enterWindow ==
        (WinHandle)FrmGetFormPtr(frmMain)) {
      debug(("winEnterEvent for main window"));
      if (update_deferred) {
        debug(("processing deferred update"));
        DrawPuzzle(fe, update_deferred);
        update_deferred = 0;
      }
      entered = 1;
    }
    break;

  case winDisplayChangedEvent:
#ifdef HAVE_HANDERA_SDK
  case displayExtentChangedEvent:
#endif
    {
      RectangleType newbounds;
      WinHandle frmWH, actWH;

      debug(("Display size has changed."));
      frm = FrmGetActiveForm();
      frmWH = FrmGetWindowHandle(frm);
      actWH = WinGetActiveWindow();
      if (frmWH == actWH) {
        /* Do not redraw when active window is not the form window,
         * e.g. when a menu is displayed
         */
        FixupWinGetBounds(WinGetDisplayWindow(), &newbounds);
        drect("New size", &newbounds);
        WinSetBounds(frmWH, &newbounds);

        app_new_size(fe, &newbounds, 0);

        DrawPuzzle(fe, frmRedrawUpdateCode);
      }
      handled = true;
    }
    break;

  case nilEvent:
    if (fe->flags & FF_TIMERON) {
      UInt32 tt_now = TimGetTicks();
      float elapsed = (tt_now - fe->tt) / (float)tps;
      debug(("nilEvent (for TIMERON)"));
      debug(("tt_then %lu, tt_now %lu, elapsed %lu ticks (%g seconds)",
            fe->tt, tt_now, (tt_now - fe->tt), (double)elapsed));
      fe->tt = tt_now;
      midend_timer(fe_me(fe), elapsed);
      handled = true;
    }
    if (fe->flags & FF_DOWNPENDING) {
      /* if we've held the pen down for long enough without moving
       * it, signal that it (the right button) has actually been pressed. */
      debug(("nilEvent (for DOWNPENDING)"));
      debug(("tt = %lu, down_tt = %lu, ISRIGHT_TICKS=%lu",
             TimGetTicks(), fe->down_tt, ISRIGHT_TICKS));
      if (ISRIGHT(fe)) {
        debug(("nilEvent decides ISRIGHT"));
        fe->flags |= FF_ISRIGHT;
        MainFormPenDown(fe);
      }
      handled = true;
    }
    /*if (!handled) debug(("nilEvent (unwanted)"));*/
    break;

  default:
    break;
  }

  return (handled);
}

/*static UInt16 EasyListGetSelection(FormType *frm, UInt16 listid)
{
  ListType *list = (ListType *)FrmGetObjectPtr(frm,
                               FrmGetObjectIndex(frm, listid));
  return LstGetSelection(list);
}

static void EasyListSetSelection(FormType *frm, UInt16 listid, UInt16 sel)
{
  ListType *list = (ListType *)FrmGetObjectPtr(frm,
                               FrmGetObjectIndex(frm, listid));
  LstSetSelection(list, sel);
}*/

config_item *palm_get_config(frontend *fe, int which, char **wintitle)
{
  config_item *ret = NULL;
  char *titlebuf, buf[40];

  titlebuf = snewn(40, char);

  switch (which) {
  case CFG_PALM_RCLICK:
    sprintf(titlebuf, "Right-click tap conversion");
    *wintitle = titlebuf;

    ret = snewn(3, config_item);
    ret[0].name = "Hold delay (ms)";
    ret[0].type = C_STRING;
    sprintf(buf, "%d", options->rclick_millis);
    ret[0].sval = dupstr(buf);
    ret[0].ival = 0;

    ret[1].name = "Drag to cancel distance (pix)";
    ret[1].type = C_STRING;
    sprintf(buf, "%d", options->rclick_offset);
    ret[1].sval = dupstr(buf);
    ret[1].ival = 0;

    ret[2].name = NULL;
    ret[2].type = C_END;
    ret[2].sval = NULL;
    ret[2].ival = 0;

    break;

  default:
    break;
  }
  return ret;
}

char *palm_set_config(frontend *fe, int which, config_item *cfg)
{
  char *ret;

  switch (which) {
  case CFG_PALM_RCLICK:
    options->rclick_millis = atoi(cfg[0].sval);
    options->rclick_offset = atoi(cfg[1].sval);
    return CheckOptions(options);

  default:
    ret = "Unknown CFG_PALM type";
  }
  if (!ret) palm_free_config(fe, which, cfg);
  return ret;
}

void palm_free_config(frontend *fe, int which, config_item *cfg)
{
  switch (which) {
  case CFG_PALM_RCLICK:
    break;

  default:
    assert(!"shouldn't get here");
  }
}

/* --- Generic application event handling --- */

static Boolean AppHandleEvent( EventPtr eventP)
{
  UInt16 formId;
  FormPtr frmP;

  switch (eventP->eType) {
  case frmLoadEvent:
    formId = eventP->data.frmLoad.formID;

    /* Set the event handler for the form.  The handler of the currently
       active form is called by FrmHandleEvent each time is receives an
       event. */
    switch (formId)
    {
    case frmMain:
      frmP = FrmInitForm(formId);
      FrmSetActiveForm(frmP);
      FrmSetEventHandler(frmP, MainFormHandleEvent);
      SetupDIA(frmP);
      break;

    case frmSelect:
      frmP = FrmInitForm(formId);
      FrmSetActiveForm(frmP);
      FrmSetEventHandler(frmP, ListHandleEvent);
      break;

    default:
      ErrNonFatalDisplay("Invalid Form Load Event");
      break;

    }
    return true;
    break;

  default:
    break;
  }

  return false;
}

static int AppEventLoop(void)
{
  UInt16 error;
  EventType event, eventToAdd;
  FormType* frmP;

  /* XXX why don't we check error in this function? */
  do {
    EvtGetEvent(&event, (fe_global.flags & (FF_TIMERON|FF_DOWNPENDING))
                ? (tps/50) : evtWaitForever);

    /* Check for WinEnterEvent */
    if ((event.eType == winEnterEvent) && dia_features) {
      frmP = FrmGetActiveForm();
      if (frmP) {
        MemSet(&eventToAdd, sizeof(EventType), 0);
        eventToAdd.eType = (eventsEnum) winDisplayChangedEvent;
        EvtAddUniqueEventToQueue(&eventToAdd, 0, true);
      }
    }

    if (!SysHandleEvent(&event))
      if (!MenuHandleEvent(0, &event, &error))
        if (!AppHandleEvent(&event))
          FrmDispatchEvent(&event);

    /* (from example code...) */
    /* Check the heaps after each event */
#if EMULATION_LEVEL != EMULATION_NONE
    MemHeapCheck(0);
    MemHeapCheck(1);
#endif

  } while (event.eType != appStopEvent);
  return 0;
}

static void AppGetOSCapabilities(void)
{
  UInt16 err;
  UInt32 version, attr;

  hd_features = 0; dia_features = 0; scr_density = 1.0;
  err = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &version);
  if (version >= 4) {
    hd_features = 1;
    WinScreenGetAttribute(winScreenDensity, &attr);
    scr_density = (double)attr / (double)kDensityLow;
    kdensity = attr;
  }
  handera = clie = 0;

  if (!hd_features) {
#ifdef HAVE_HANDERA_SDK
    if (FtrGet(TRGSysFtrID, TRGVgaFtrNum, &version) == 0) {
      /* this is a HandEra, with a 240x240 screen. */
      VgaSetScreenMode(screenMode1To1, rotateModeNone);
      handera = 1;
      debug(("handera detected."));
    }
#endif

#ifdef HAVE_CLIE_SDK
    if ((err = SysLibFind(sonySysLibNameHR, &hrRefNum))) {
      if (err == sysErrLibNotFound){
        /* couldn't find lib, try to load it */
        err = SysLibLoad('libr', sonySysFileCHRLib, &hrRefNum);
      }
    }
    if (!err) {
      /* Executes Open library. */
      err = HROpen(hrRefNum);
      if (!err) {
        UInt32 screen_width = hrWidth;
        UInt32 screen_height = hrHeight;
        err = HRWinScreenMode(hrRefNum, winScreenModeSet,
                              &screen_width, &screen_height, NULL, NULL );
        if (err != errNone) {
          /* Screen mode remains unchanged. */
        } else {
          /* we got high-resolution mode */
          clie = 1;
          scr_density = 2.0;
          debug(("clie detected."));
        }
      }
    }
#endif
    ;
  }
  debug(("scr_density = %f", scr_density));

  tps = SysTicksPerSecond();
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romv);

  err = FtrGet(pinCreator, pinFtrAPIVersion, &version);
  if (!err && version) {
    dia_features = 1;
    debug(("Enabling DIA features, PIN version is 0x%lx", version));
  }

  if (FrmGlueNavIsSupported()) {
    nav_features = 1;
    debug(("5way navigator supported"));
  }
}

/* --- Startup code --- */

/* BEWARE that none of this code may use globals.
 * PalmOS only guarantees globals if it starts you up with the
 * sysAppLaunchCmdNormalLaunch launch code; anything executed
 * before you check your launch code must be stand-alone. */

/* (all example code...) */
static Err RomVersionCompatible(UInt32 requiredVersion,
                                UInt16 launchFlags)
{
  UInt32 romVersion;

  /* See if we're on in minimum required version of the ROM or later. */
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
  if (romVersion < requiredVersion)
  {
    if ((launchFlags & (sysAppLaunchFlagNewGlobals |
                        sysAppLaunchFlagUIApp)) ==
        (sysAppLaunchFlagNewGlobals |
         sysAppLaunchFlagUIApp))
    {
      FrmAlert (altROMVersion);

      /* Pilot 1.0 will continuously relaunch this app unless we
       * switch to another safe one. */
      if (romVersion < 0x02000000)
      {
        AppLaunchWithCommand(sysFileCDefaultApp,
                             sysAppLaunchCmdNormalLaunch, NULL);
      }
    }

    return (sysErrRomIncompatible);
  }

  return (0);
}

UInt32 PilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
  UInt16 error;
  UInt32 version35;
  int ret = 0, caught = 0;

  version35 = sysMakeROMVersion(3,5,0,0,0);
  if ((error = RomVersionCompatible(version35, launchFlags)))
    return error;

  switch (cmd) {
  case sysAppLaunchCmdNormalLaunch:
#ifdef DEBUGGING
    if (HostGetHostPlatform() != hostPlatformPalmOS)
      HostTraceInit();
#endif
    AppGetOSCapabilities();
    ErrTry {
      if ((error = StartApplication()))
        return error;
      ret = AppEventLoop();
    }
    ErrCatch (exc_err) {
      error = exc_err;
      caught = 1;
      debug(("Caught OOM exception %d", error));
    } ErrEndCatch;
    if (caught) FrmAlert(noMemory);
    StopApplication(ret);
#ifdef DEBUGGING
    if (HostGetHostPlatform() != hostPlatformPalmOS)
      HostTraceClose();
#endif
    break;

  case sysAppLaunchCmdNotify:
    if (((SysNotifyParamType*) cmdPBP)->notifyType ==
        sysNotifyDisplayResizedEvent)
    {
      /* DIA size-changed notify; add a windowy notification
       * to the event queue. */
      EventType resizedEvent;
      MemSet(&resizedEvent, sizeof(EventType), 0);
      resizedEvent.eType = winDisplayChangedEvent;
      EvtAddUniqueEventToQueue(&resizedEvent, 0, true);
    }
    break;


  default:
    break;
  }

  return (0);
}
