#include "palm-extras.h"
#include "palmrcp.h"
#include <PalmOSGlue.h>

static Int16 TopVisibleGame = 0;
static Int16 SelectedGame = 0;

static void HighlightItem(RectanglePtr rP) CUSTOM_SECT;
static void HighlightItem(RectanglePtr rP)
{
  WinPushDrawState();
  WinSetDrawMode(winSwap);
  WinSetPatternType (blackPattern);
  WinSetBackColor(UIColorGetTableEntryIndex(UIFieldBackground));
  WinSetForeColor(UIColorGetTableEntryIndex(UIObjectSelectedFill));
  WinPaintRectangle(rP, 0);

  if (UIColorGetTableEntryIndex(UIObjectSelectedFill) != UIColorGetTableEntryIndex(UIObjectForeground)) {
    WinSetBackColor(UIColorGetTableEntryIndex(UIObjectForeground));
    WinSetForeColor(UIColorGetTableEntryIndex(UIObjectSelectedForeground));
    WinPaintRectangle(rP, 0);
  }
  WinPopDrawState();
}

/* bitmap drawing:
 *
 * (from http://palmosprogramming.blogspot.com/2007/08/drawing-bitmap-family-member.html)
 *
 * MemHandle hRsc = DmGetResource(bitmapRsc, TestBitmapFamily);
 * BitmapType* bitmapP = (BitmapType*) MemHandleLock(hRsc);
 * WinDrawBitmap(bitmapP, topLeftCoordX, topLeftCoordY);
 * MemHandleUnlock(hRsc);
 * DmReleaseResource(hRsc);
 */

static void ListDrawGame(void *tblP, Int16 row, Int16 col, RectanglePtr bounds) CUSTOM_SECT;
static void ListDrawGame(void *tblP, Int16 row, Int16 col, RectanglePtr bounds)
{
  UInt16 len;
  char *gamename;
  Int16 x, y, width;

  gamename = (char *)TblGetItemPtr(tblP, row, col);
  FntSetFont(stdFont);
  x = bounds->topLeft.x + 1;
  y = bounds->topLeft.y;
  width = bounds->extent.x -1;

  len = StrLen(gamename);
  if (FntWidthToOffset(gamename, len, width, NULL, NULL) == len)
    WinDrawChars(gamename, len, x, y);
  else {
    Int16 namewidth;
    len = FntWidthToOffset(gamename, len, width - FntCharWidth(chrEllipsis),
             NULL, &namewidth);
    WinDrawChars(gamename, len, x, y);
    WinDrawChar(chrEllipsis, x + namewidth, y);
  }
  if (TblGetItemInt(tblP, row, col) == SelectedGame)
    HighlightItem(bounds);
}

static UInt16 ListNumberOfRows(TablePtr table) CUSTOM_SECT;
static UInt16 ListNumberOfRows(TablePtr table)
{
  UInt16 rows, rowsInTable, tblheight;
  FontID currfont;
  RectangleType r;

  rowsInTable = TblGetNumberOfRows(table);
  TblGetBounds(table, &r);
  tblheight = r.extent.y;

#ifdef HAVE_HANDERA_SDK
  if (handera)
    currfont = FntSetFont(VgaBaseToVgaFont(stdFont));
  else
#endif
    currfont = FntSetFont(stdFont);
  rows = tblheight / FntLineHeight();
  FntSetFont(currfont);

  return (rows <= rowsInTable ? rows : rowsInTable);
}

static void ListUpdateScrollBar(FormPtr frmP) CUSTOM_SECT;
static void ListUpdateScrollBar(FormPtr frmP)
{
  UInt16 pos, max, rows, cols;
  TablePtr tblP;

  tblP = FrmEasyPtr(frmP, frmSelectTable);
  rows = ListNumberOfRows(tblP);
  cols = TblGlueGetNumberOfColumns(tblP);
  if ((gamecount + cols - 1) / cols > rows) {
    pos = (UInt16)TblGetItemInt(tblP, 0, 0) / cols;
    max = (gamecount + cols - 1) / cols - rows;
  }
  else {
    pos = max = 0;
  }
  SclSetScrollBar(FrmEasyPtr(frmP, frmSelectScrollBar), pos, 0, max, rows);
}

static void ListLoadTable(FormPtr frmP) CUSTOM_SECT;
static void ListLoadTable(FormPtr frmP)
{
  UInt16 row, col, gameno, nrows, ncols;
  UInt16 lineheight, dataheight, tblheight;
  FontID currfont;
  TablePtr tblP;
  RectangleType r;

  tblP = FrmEasyPtr(frmP, frmSelectTable);
  TblGetBounds(tblP, &r);
  tblheight = r.extent.y;

#ifdef HAVE_HANDERA_SDK
  if (handera)
    currfont = FntSetFont(VgaBaseToVgaFont(stdFont));
  else
#endif
    currfont = FntSetFont(stdFont);
  lineheight = FntLineHeight();
  FntSetFont(currfont);

  dataheight = 0;
  gameno = TopVisibleGame;

  /* For each element in the table store the game number in the table item
   * that will display the game title */
  nrows = TblGetNumberOfRows(tblP);
  ncols = TblGlueGetNumberOfColumns(tblP);
  for (row = 0;  row < nrows; row++) {
    if (tblheight >= dataheight + lineheight) {
      for (col = 0; col < ncols; col++, gameno++) {
        TblSetItemStyle(tblP, row, col, customTableItem);
        TblSetItemFont(tblP, row, col, stdFont);
        if (gameno < gamecount) {
          TblSetItemInt(tblP, row, col, gameno);
          TblSetItemPtr(tblP, row, col, (void *)gamelist[gameno]->name);
        }
        else {
          TblSetItemInt(tblP, row, col, -1);
          TblSetItemPtr(tblP, row, col, "");
        }
      }
      TblSetRowHeight(tblP, row, lineheight);
      TblSetRowUsable(tblP, row, gameno < (gamecount + ncols));
      TblMarkRowInvalid(tblP, row);
      dataheight += lineheight;
    }
    else {
      TblSetRowHeight(tblP, row, lineheight);
      TblSetRowUsable(tblP, row, false);
    }
  }

  /* Update scroll bar */
  ListUpdateScrollBar(frmP);
}

static void MoveSelection(Direction dir) CUSTOM_SECT;
static void MoveSelection(Direction dir)
{
  Int16 oldcol, newcol;
  FormPtr frmP = FrmGetActiveForm();
  TablePtr tblP = FrmEasyPtr(frmP, frmSelectTable);
  UInt16 rows = ListNumberOfRows(tblP);
  UInt16 ncols = TblGlueGetNumberOfColumns(tblP);

  switch (dir) {
  case Left:
  case Right:
    SelectedGame += (dir == Left ? -1 : 1);
    /* wrap around */
    if (SelectedGame < 0) SelectedGame = gamecount - 1;
    if (SelectedGame >= gamecount) SelectedGame = 0;
    break;

  case Up:
  case Down:
    oldcol = SelectedGame % ncols;
    if (dir == Up) {
      newcol = oldcol - 1;
      if (newcol < 0) newcol = ncols - 1;
      SelectedGame -= ncols;
      if (SelectedGame < 0) {
        SelectedGame = gamecount / ncols * ncols + newcol;
        if (SelectedGame >= gamecount) SelectedGame -= ncols;
      }
    }
    else {
      /* dir == Down */
      newcol = oldcol + 1;
      if (newcol >= ncols) newcol = 0;
      SelectedGame += ncols;
      if (SelectedGame >= gamecount) {
        SelectedGame = newcol;
      }
    }
    break;

    default:
      return;
  }

  /* Make sure the selected game stays visible */
  if (SelectedGame < TopVisibleGame)
    TopVisibleGame = SelectedGame - SelectedGame % ncols;
  else if (SelectedGame > TopVisibleGame + rows*ncols - 1) {
    TopVisibleGame = (SelectedGame - rows*ncols + ncols);
    TopVisibleGame -= TopVisibleGame % ncols;
  }

  ListLoadTable(frmP);
  TblRedrawTable(tblP);
}

static void ListScroll(Int16 lines) CUSTOM_SECT;
static void ListScroll(Int16 lines)
{
  Int16 i;
  UInt16 rows, lastrow, sclamount;
  TablePtr tblP;
  RectangleType r, vacated;
  WinDirectionType dir;
  FormPtr frmP = FrmGetActiveForm();

  tblP = FrmEasyPtr(frmP, frmSelectTable);
  TopVisibleGame = max(TopVisibleGame + lines*2, 0);
  TopVisibleGame = min(TopVisibleGame, (gamecount - 1) & ~1);

  /* Move the bits that will remain visible */
  rows = ListNumberOfRows(tblP);
  if (((lines > 0) && (lines < rows)) ||
      ((lines < 0) && (-lines < rows))) {
    sclamount = 0;
    if (lines > 0) {
      lastrow = TblGetLastUsableRow(tblP) - 1;
      for (i = 0; i < lines; i++) {
        sclamount += TblGetRowHeight(tblP, lastrow);
        TblRemoveRow(tblP, 0);
      }
      dir = winUp;
    }
    else {
      for (i = 0; i < -lines; i++) {
        sclamount += TblGetRowHeight(tblP, 0);
        TblInsertRow(tblP, 0);
      }
      dir = winDown;
    }
    TblGetBounds(tblP, &r);
    WinScrollRectangle(&r, dir, sclamount, &vacated);
    WinEraseRectangle(&vacated, 0);
  }

  ListLoadTable(frmP);
  TblRedrawTable(tblP);
}

static void ListInit(FormPtr frmP) CUSTOM_SECT;
static void ListInit(FormPtr frmP)
{
  TablePtr tblP;
  UInt16 gameno, nrows, ncols, i;

  SelectedGame = fe_global.currgame;

  tblP = FrmEasyPtr(frmP, frmSelectTable);
  nrows = ListNumberOfRows(tblP);
  ncols = TblGlueGetNumberOfColumns(tblP);

  /* Is the current game before the first visible game ? */
  if (TopVisibleGame > SelectedGame)
    TopVisibleGame = SelectedGame - SelectedGame % ncols;
  else {
    /* Is the current game after the last visible game ? */
    gameno = TopVisibleGame;
    gameno = min(gamecount - 1, gameno + nrows*ncols - 1);
    if (gameno < SelectedGame)
      TopVisibleGame = gameno - gameno % ncols;
  }

  /* Make sure we show a full display of games */
  TopVisibleGame = min(TopVisibleGame,
                       ((gamecount + ncols - 1) / ncols - nrows) * ncols);
  TopVisibleGame = max(TopVisibleGame, 0);
  ListLoadTable(frmP);

  TblHasScrollBar(tblP, true);
  for (i=0; i < ncols; i++) {
    TblSetCustomDrawProcedure(tblP, i, ListDrawGame);
    TblSetColumnUsable(tblP, i, true);
  }
}

Boolean ListHandleEvent(EventPtr event)
{
  FormPtr frmP;
  Direction dir;
  Boolean handled = false;

  switch (event->eType) {
  case frmOpenEvent:
    frmP = FrmGetActiveForm();
#ifdef HAVE_HANDERA_SDK
    if (handera) VgaFormModify(frmP, vgaFormModify160To240);
#endif
    ListInit(frmP);
    if (nav_features == 1) {
      FrmGlueNavRemoveFocusRing(frmP);
      FrmSetNavState(frmP, 0);
    }
    FrmDrawForm(frmP);
    handled = true;
    break;

  case frmObjectFocusTakeEvent:
    /* Eat this event, we do not want any focus rings */
    handled = true;
    break;

  case ctlSelectEvent:
    if (event->data.ctlSelect.controlID == frmSelectCancel) {
      FrmGotoForm(frmMain);
      handled = true;
    }
    break;

  case sclRepeatEvent:
    ListScroll(event->data.sclRepeat.newValue - event->data.sclRepeat.value);
    break;

  case keyDownEvent:
    debug(("ListHandleEvent key: chr = %04x, code = %04x, mod = %04x", event->data.keyDown.chr, event->data.keyDown.keyCode, event->data.keyDown.modifiers));

    if (fiveway_key(event, &dir)) {
      if (dir == Center) {
        if (SelectedGame != fe_global.currgame && SelectedGame >= 0) {
          /* Reset the draw window to full screen, since app_change_game
           * uses WinGetDrawWindowBounds to compute the window size. Right
           * now the modal form frame is outside the draw window */
          WinSetDrawWindow(WinGetDisplayWindow());
          app_change_game(&fe_global, SelectedGame);
        }
        FrmGotoForm(frmMain);
      }
      else
        MoveSelection(dir);
      handled = true;
    }
    break;

  case tblSelectEvent:
    SelectedGame = TblGetItemInt(event->data.tblSelect.pTable,
				 event->data.tblSelect.row,
				 event->data.tblSelect.column);
    if (SelectedGame != fe_global.currgame && SelectedGame >= 0) {
      ListLoadTable(FrmGetActiveForm());
      TblRedrawTable(event->data.tblSelect.pTable);
      /* Reset the draw window to full screen, since app_change_game
       * uses WinGetDrawWindowBounds to compute the window size. Right
       * now the modal form frame is outside the draw window */
      WinSetDrawWindow(WinGetDisplayWindow());
      app_change_game(&fe_global, SelectedGame);
    }
    FrmGotoForm(frmMain);
    handled = true;
    break;

  default:
  }
  return handled;
}
