#! /usr/bin/env python

import os
import sys
import re

def die(s):
  print >>sys.stderr, s
  sys.exit(1)

def usage():
  die("Usage: static-sect.py MAKEDEF <sections> | VERSION <version> | <source file>")

def mksect(fstem):
  return "g_%s" % (fstem[0:6])

def exceptions(sfile):
  sections = open("sections.conf")
  for l in sections:
    l = l.strip()
    if l == "" or l.startswith('#'): continue
    if not l.startswith(sfile): continue

    attrs = re.sub(r'\s+', ' ', l).split(' ')

    sects = attrs[1]
    if sects.count('/') == 0:
      sect = mksect(sects)
      sect2 = ""
    else:
      (sect,sect2) = map(mksect, sects.split('/'))

    if len(attrs) == 3:
      fns = attrs[2].split(',')
    else:
      fns = []

    return (sect, sect2, fns)

  return (mksect(sfile), "", [])

if len(sys.argv) < 1: usage()
sfile = sys.argv[1]

if sfile == 'MAKEDEF':
  dsect = {}
  for sect in sys.argv[2:]:
    (s1, s2, ignored) = exceptions(os.path.basename(sect))
    dsect[s1] = 1
    dsect[s2] = 1
  print "application { \"Puzzles\" SGTp stack=16384 }"
  print "multiple code { ",
  for sect in dsect.keys():
    print "%s " % (sect),
  print "}"
  sys.exit(0)

if sfile == 'VERSION':
  ver = sys.argv[2]
  if ver.startswith('-DREVISION'):
    (ver,nsub) = re.subn(r"^-DREVISION=(\d+)-palm-(\d+)", r'"\1p\2"', ver)
  else:
    (ver,nsub) = re.subn(r"^(\d+)$", r'"\1p-devel"', ver)
  if nsub == 0: die("Unknown version $version")

  rcpin = open("palm.rcp.in")
  for l in rcpin:
    l = re.sub(r'^VERSION\s".+"$', "VERSION %s" % (ver), l)
    print l,
  rcpin.close()
  sys.exit(0)

if sfile == 'LIST':
  pfiles = sys.argv[2:]
  pnames = []
  puzzre = re.compile('^#define thegame (\S+)')
  for p in pfiles:
    psrc = open(p)
    for l in psrc:
      puzzmo = puzzre.match(l)
      if puzzmo: pnames.append(puzzmo.group(1))

  print "/* THIS FILE IS AUTO-GENERATED! */"
  print "#include \"puzzles.h\""
  print
  for puzz in pnames:
    print "extern const game %s;" % (puzz)
  print
  print "CONST_GLOBAL game *gamelist[] = {"
  for puzz in pnames:
    print "    (CONST_GLOBAL game*)&%s," % (puzz)
  print "};"
  print
  print "CONST_GLOBAL int gamecount = %d;" % (len(pnames))
  sys.exit(0)

(sfilebase, sfileext) = os.path.splitext(os.path.basename(sfile))
(sect, sect2, s2fns) = exceptions(sfilebase)

ip = open(sfile)
lines = ip.read() # read whole file at once.
ip.close()

def fixup(mo):
  (decl, post) = mo.group(1,2)

  # if our declaration doesn't have preprocessor guff,
  # put it on one line. If it does, we lose (and will smash __LINE__).
  decl_oneline = decl
  if decl.count('#') == 0:
    decl_oneline = re.sub(r'\s+', ' ', decl_oneline)

  dsect = sect
  for fn in s2fns:
    if decl_oneline.count(fn) > 0: dsect = sect2

  ret = "%s __attribute__ ((section (\"%s\")));" % (decl_oneline, dsect)
  if post != ';':
    ret = ret + " %s\n{" % (decl)
  print >>sys.stderr, "%s:\t%s..." % (dsect, ret[0:50])
  return ret

def debug(mo):
  print >>sys.stderr, "debug: %s" % (mo.groups(0))

if sfileext == '.h':
  flines = re.sub(r'(?m)^(\w+[^({=]+?\([^)]+\))(\s+{|;)', fixup, lines)
else:
  flines = re.sub(r'(?m)^(static[^({=]+?\([^)]+\))(\s+{|;)', fixup, lines)

print flines

