/* trifle.c - fill a triangle */

/* Hurray for google. Found this code at:
 * d0sk38$1d9$1$830fa79f@news.demon.co.uk and
 * d1254h$4ne$1$8300dec7@news.demon.co.uk (correction)
 */

/*
 * Copyright Roger Willcocks, 2005
 * Permission is granted to copy and distribute this
 * work provided that this notice remains intact.
 * rkww - 11 Mar 2005
 */

/* Hacked around (for PalmOS) and tidied up a bit by
 * James Harvey 28 Apr 2005 */

/* XXX We can change the 'long long' to 'int' in initedge() assuming
 * we're happy that it'll only work for sizes < 2<<16. I dunno if
 * prc-tools/palmos does 'long long' properly. */

#include "palm-extras.h"

typedef struct _vertex {
    int x, y;
} vertex;

typedef struct _triangle {
    vertex v[3];
} triangle;

static void span(int line, int x0, int x1)
{
  WinDrawLine((Coord)x0, (Coord)line, (Coord)x1, (Coord)line);
}

/* swap vertices p & q if necessary so y[p] <= y[q] */

static int swapv(triangle* t, int p, int q) DRAW_SECT;
static int swapv(triangle* t, int p, int q)
{
    if (t->v[p].y > t->v[q].y) {
        vertex v = t->v[p];
        t->v[p] = t->v[q];
        t->v[q] = v;
        return 1;
    }
    return 0;
}

/*
 * Suppose a rectangular grid. Vertices lay at the centre
 * of grid rectangles. We 'cut' the plane along grid lines.
 * Life is made simpler by doubling the coordinates.
 */

typedef struct _edge {
    int xint, xfrac;
    int dxint, dxfrac;
    int dy, life;
} edge;

/*static void debugedge(const char *s, edge *e) DRAW_SECT;
static void debugedge(const char *s, edge *e)
{
  debug(("edge %s: x %d-%d, dx %d-%d, dy %d, life %d",
         s, e->xint, e->xfrac, e->dxint, e->dxfrac, e->dy, e->life));
}*/

/* initialize DDA for edge p->q */

static edge initedge(triangle* t, int p, int q) DRAW_SECT;
static edge initedge(triangle* t, int p, int q)
{
    long long x; /* need long long intermediate for muldivmod */
    int dx, dy;

    edge e;

    /*debug(("initedge: edge from (%d,%d) to (%d,%d)",
           t->v[p].x, t->v[p].y, t->v[q].x, t->v[q].y));*/

    dx = t->v[q].x - t->v[p].x;
    dy = t->v[q].y - t->v[p].y;

    MemSet(&e, sizeof(edge), 0); /* WHY!!! WHY CHANGE ARGUMENT ORDER!!! */
    e.life = dy;
    e.dy = dy += dy;

    if (dy == 0)
        return e;

    /* initial intersection is x + (dx/dy) / 2 */

    x = t->v[p].x * (long long)e.dy + dx;
    if (dx < 0) x += e.life;

    e.xint = (int)(x / dy);
    e.xfrac = (int)(x % dy);

    /* -3.5 = -4 + .5 */

    if (e.xfrac < 0) {
        e.xint--;
        e.xfrac += dy;
    }

    dx += dx;
    e.dxint = dx / dy;
    e.dxfrac = dx % dy;

    if (e.dxfrac < 0) {
        e.dxint--;
        e.dxfrac += dy;
    }
    return e;
}

/* advance x position by dx/dy avoiding division in this inner loop */

static void advance(edge* e) DRAW_SECT;
static void advance(edge* e)
{
    e->xint += e->dxint;
    if ((e->xfrac += e->dxfrac) >= e->dy) {
        e->xfrac -= e->dy;
        e->xint++;
    }
    e->life--;
}

/*
static void debugtriangle(triangle *t) DRAW_SECT;
static void debugtriangle(triangle *t)
{
  debug(("filltri: (%d,%d) --> (%d,%d) --> (%d,%d)",
         t->v[0].x, t->v[0].y,
         t->v[1].x, t->v[1].y,
         t->v[2].x, t->v[2].y));
}*/

static void filltriangle(triangle* t) DRAW_SECT;
static void filltriangle(triangle* t)
{
    edge shortedge, longedge;
    int line;

    /* switch vertices around so they're in y order; overwrites
       source triangle. Bubble sort :-) */

    while (swapv(t, 0, 1) + swapv(t, 1, 2))
        ;
    /*debugtriangle(t);*/

    /* set up DDA for first two edges */

    longedge = initedge(t, 0, 2);
    shortedge = initedge(t, 0, 1);

    /* draw upper triangle */

    /*debug(("Upper triangle: "));
    debugedge("longedge", &longedge);
    debugedge("shortedge", &shortedge);*/

    line = t->v[0].y;

    while (shortedge.life) {
        span(line++, longedge.xint, shortedge.xint);
        advance(&longedge);
        advance(&shortedge);
    }

    /* now set up DDA for third edge */

    shortedge = initedge(t, 1, 2);

    /* and draw lower triangle */

    /*debug(("Lower triangle: "));
    debugedge("longedge (remaining)", &longedge);
    debugedge("shortedge", &shortedge);*/

    while (longedge.life) {
        span(line++, longedge.xint, shortedge.xint);
        advance(&longedge);
        advance(&shortedge);
    }
}

void WinFillTriangle(Coord x0, Coord y0,
                     Coord x1, Coord y1,
                     Coord x2, Coord y2)
{
  triangle t;

  t.v[0].x = (int)x0; t.v[0].y = (int)y0;
  t.v[1].x = (int)x1; t.v[1].y = (int)y1;
  t.v[2].x = (int)x2; t.v[2].y = (int)y2;
  filltriangle(&t);
}

/* points[i*2+0] is x coord, points[i*2+1] is y coord. */
void WinFillPolygon(int *points, int npoints, float xsc, float ysc)
{
  triangle t;
  int i;

  switch (npoints) {
  case 0:
    return;

  case 1:
    WinDrawPixel(points[0]*xsc, points[1]*ysc);
    return;

  case 2:
    WinDrawLine(points[0]*xsc, points[1]*ysc,
                points[2]*xsc, points[3]*ysc);
    return;

  default:
  }

/*#ifdef DEBUGGING
  for (i = 0; i < npoints; i++) {
    debug(("polygon point #%d: (%d, %d)",
           i, points[i*2+0], points[i*2+1]));
  }
#endif*/

  for (i=2; i < npoints; i++) {
    /* have to fill these all in each time, because filltriangle()
     * might be swapping them around under our feet. */
    t.v[0].x = points[0]*xsc;
    t.v[0].y = points[1]*ysc;

    t.v[1].x = points[(i-1)*2+0]*xsc;
    t.v[1].y = points[(i-1)*2+1]*ysc;

    t.v[2].x = points[i*2+0]*xsc;
    t.v[2].y = points[i*2+1]*ysc;

    filltriangle(&t);
  }
}
