/*
 *  linux/lib/vsprintf.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

/* vsprintf.c -- Lars Wirzenius & Linus Torvalds. */
/*
 * Wirzenius wrote this portably, Torvalds fucked it up :-)
 */

/* Adapted for the use in the 332 kernel, then again for PalmOS/Pilot
 * Added floating point support, D. Jeff Dionne, 1995, 1997
 * Added %g support (using printDouble), jrh 2005
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "palm-extras.h"

/* we use this so that we can do without the ctype library */
#define is_digit(c)	((c) >= '0' && (c) <= '9')

static int skip_atoi(const char **s)
{
	int i=0;

	while (is_digit(**s))
		i = i*10 + *((*s)++) - '0';
	return i;
}

static size_t strnlen(const char * s, size_t count)
{
  const char *sc;
  
  for (sc = s; count-- && *sc != '\0'; ++sc)
    /* nothing */;
  return sc - s;
}

#define ZEROPAD	1		/* pad with zero */
#define SIGN	2		/* unsigned/signed long */
#define PLUS	4		/* show plus */
#define SPACE	8		/* space if plus */
#define LEFT	16		/* left justified */
#define SPECIAL	32		/* 0x */
#define LARGE	64		/* use 'ABCDEF' instead of 'abcdef' */

#define do_div(n,base) ({ \
int __res; \
__res = ((unsigned long) n) % (unsigned) base; \
n = ((unsigned long) n) / (unsigned) base; \
__res; })

#define OFLOWCHECK do { if (strmax && (str >= strmax)) { goto overflow; } } while(0)

#define NCHAR(c) do { *str++ = (c); OFLOWCHECK; } while(0)

static char * number(char *str, char *strmax, long num, int base, int size, int precision
	,int type)
{
	char c,sign,tmp[66];
	const char *digits="0123456789abcdefghijklmnopqrstuvwxyz";
	int i;

	if (type & LARGE)
		digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	if (type & LEFT)
		type &= ~ZEROPAD;
	if (base < 2 || base > 36)
		return 0;
	c = (type & ZEROPAD) ? '0' : ' ';
	sign = 0;
	if (type & SIGN) {
		if (num < 0) {
			sign = '-';
			num = -num;
			size--;
		} else if (type & PLUS) {
			sign = '+';
			size--;
		} else if (type & SPACE) {
			sign = ' ';
			size--;
		}
	}
	if (type & SPECIAL) {
		if (base == 16)
			size -= 2;
		else if (base == 8)
			size--;
	}
	i = 0;
	if (num == 0)
		tmp[i++]='0';
	else while (num != 0)
		tmp[i++] = digits[do_div(num,base)];
	if (i > precision)
		precision = i;
	size -= precision;
	if (!(type&(ZEROPAD+LEFT)))
		while(size-->0)
			NCHAR(' ');
	if (sign)
		NCHAR(sign);
	if (type & SPECIAL) {
		if (base==8) {
			NCHAR('0');
                } else if (base==16) {
			NCHAR('0');
			NCHAR(digits[33]);
		}
        }
	if (!(type & LEFT))
		while (size-- > 0)
			NCHAR(c);
	while (i < precision--)
		NCHAR('0');
	while (i-- > 0)
		NCHAR(tmp[i]);
	while (size-- > 0)
		NCHAR(' ');
overflow:
	return str;
}

static int vsnprintf_int(char *buf, size_t slen, const char *fmt, va_list args)
{
	int len;
	unsigned long num;
	int i, base;
	char *str = buf, *strmax = (slen) ? buf+slen-1 : NULL;
	char *s;
#ifdef PRINT_FLOATS
	float fl;
#endif /* PRINT_FLOATS */
	int flags;		/* flags to number() */

	int field_width;	/* width of output field */
	int precision;		/* min. # of digits for integers; max
				   number of chars for from string */
	int qualifier;		/* 'h', 'l', or 'L' for integer fields */

	for (; *fmt ; ++fmt) {
		if (*fmt != '%') {
                        NCHAR(*fmt);
			continue;
		}
			
		/* process flags */
		flags = 0;
		repeat:
			++fmt;		/* this also skips first '%' */
			switch (*fmt) {
				case '-': flags |= LEFT; goto repeat;
				case '+': flags |= PLUS; goto repeat;
				case ' ': flags |= SPACE; goto repeat;
				case '#': flags |= SPECIAL; goto repeat;
				case '0': flags |= ZEROPAD; goto repeat;
				}
		
		/* get field width */
		field_width = -1;
		if (is_digit(*fmt))
			field_width = skip_atoi(&fmt);
		else if (*fmt == '*') {
			++fmt;
			/* it's the next argument */
			field_width = va_arg(args, int);
			if (field_width < 0) {
				field_width = -field_width;
				flags |= LEFT;
			}
		}

		/* get the precision */
		precision = -1;
		if (*fmt == '.') {
			++fmt;	
			if (is_digit(*fmt))
				precision = skip_atoi(&fmt);
			else if (*fmt == '*') {
				++fmt;
				/* it's the next argument */
				precision = va_arg(args, int);
			}
			if (precision < 0)
				precision = 0;
		}

		/* get the conversion qualifier */
		qualifier = -1;
		if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L') {
			qualifier = *fmt;
			++fmt;
		}

		/* default base */
		base = 10;

		switch (*fmt) {
		case 'c':
			if (!(flags & LEFT))
				while (--field_width > 0)
					NCHAR(' ');
			NCHAR((unsigned char) va_arg(args, int));
			while (--field_width > 0)
				NCHAR(' ');
			continue;

		case 's':
			s = va_arg(args, char *);
			if (!s)
				s = "<NULL>";

			len = strnlen(s, precision);

			if (!(flags & LEFT))
				while (len < field_width--)
					NCHAR(' ');
			for (i = 0; i < len; ++i)
				NCHAR(*s++);
			while (len < field_width--)
				NCHAR(' ');
			continue;

#ifdef PRINT_FLOATS
                case 'e':
                case 'f':
                case 'g':
                        {
                          char tmp[100];
                          fl = va_arg(args, double);
                          printDouble(tmp, fl, precision);
                          if (strmax) {
                            strncpy(str, tmp, strmax-str);
                            str[strmax-str] = '\0';
                          } else
                            strcpy(str,tmp);
                          str += strlen(str);
                          OFLOWCHECK;
                        }
                        continue;
#endif /* PRINT_FLOATS */

		case 'p':
			if (field_width == -1) {
				field_width = 2*sizeof(void *);
				flags |= ZEROPAD;
			}
                        base = 16;
                        break;

		case 'n':
			if (qualifier == 'l') {
				long * ip = va_arg(args, long *);
				*ip = (str - buf);
			} else {
				int * ip = va_arg(args, int *);
				*ip = (str - buf);
			}
			continue;

		/* integer number formats - set up the flags and "break" */
		case 'o':
			base = 8;
			break;

		case 'X':
			flags |= LARGE;
		case 'x':
			base = 16;
			break;

		case 'd':
		case 'i':
			flags |= SIGN;
		case 'u':
			break;

		default:
			if (*fmt != '%')
				NCHAR('%');
			if (*fmt)
				NCHAR(*fmt);
			else
				--fmt;
			continue;
		}
                if (*fmt == 'p')
                        num = (unsigned long) va_arg(args, void *);
                else if (qualifier == 'l')
			num = va_arg(args, unsigned long);
		else if (qualifier == 'h')
			if (flags & SIGN)
				num = va_arg(args, short);
			else
				num = va_arg(args, unsigned short);
		else if (flags & SIGN)
			num = va_arg(args, int);
		else
			num = va_arg(args, unsigned int);
		str = number(str, strmax, num, base, field_width, precision, flags);
                OFLOWCHECK;
	}
overflow:
	*str = '\0';
	return str-buf;
}

int vsnprintf(char *buf, size_t len, const char *fmt, va_list args)
{
  if (len == 0) return 0;
  else if (len == 1) {
    *buf = '\0';
    return 1;
  }
  return vsnprintf_int(buf,len,fmt,args);
}

int vsprintf(char *buf, const char *fmt, va_list args)
{
  return vsnprintf_int(buf, 0, fmt, args);
}

int snprintf(char *buf, size_t len, const char *fmt, ...)
{
  va_list args;
  int i;

  if (len == 0) return 0;
  else if (len == 1) {
    *buf = '\0';
    return 1;
  }

  va_start(args, fmt);
  i=vsnprintf_int(buf,len,fmt,args);
  va_end(args);
  return i;
}

int sprintf(char * buf, const char *fmt, ...)
{
  va_list args;
  int i;

  va_start(args, fmt);
  i=vsnprintf_int(buf,0,fmt,args);
  va_end(args);
  return i;
}
